import json
import urllib.parse as urlparse
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional
from unittest.mock import Mock

import pytest
from click import ClickException

from ggshield.cmd.auth.utils import (
    DISABLED_FLOW_MESSAGE,
    VERSION_TOO_LOW_MESSAGE,
    check_instance_has_enabled_flow,
)
from ggshield.cmd.main import cli
from ggshield.core.config import Config
from ggshield.core.oauth import (
    OAuthClient,
    OAuthError,
    get_error_param,
    get_pretty_date,
)
from tests.conftest import assert_invoke_ok

from ..utils import prepare_config


_TOKEN_RESPONSE_PAYLOAD = {
    "type": "personal_access_token",
    "account_id": 17,
    "name": "key",
    "scope": ["scan"],
    "expire_at": None,
}

_EXPECTED_URL_PARAMS = {
    "auth_mode": ["ggshield_login"],
    "client_id": ["ggshield_oauth"],
    "code_challenge_method": ["S256"],
    "response_type": ["code"],
    "scope": ["scan"],
    "utm_campaign": ["ggshield"],
    "utm_medium": ["login"],
    "utm_source": ["cli"],
}

DT_FORMAT = "%Y-%m-%dT%H:%M:%S%z"


@pytest.fixture(autouse=True)
def tmp_config(monkeypatch, tmp_path):
    monkeypatch.setattr(
        "ggshield.core.config.utils.get_auth_config_dir", lambda: str(tmp_path)
    )


class TestAuthLoginToken:
    VALID_TOKEN_PAYLOAD = {**_TOKEN_RESPONSE_PAYLOAD}
    INVALID_TOKEN_PAYLOAD = {"detail": "Invalid API key."}
    VALID_TOKEN_INVALID_SCOPE_PAYLOAD = {
        **_TOKEN_RESPONSE_PAYLOAD,
        "scope": ["read:incident", "write:incident", "share:incident", "read:member"],
    }

    @staticmethod
    def mock_autho_login_request(
        monkeypatch, status_code: int, json: Dict[str, Any]
    ) -> None:
        monkeypatch.setattr(
            "ggshield.core.client.GGClient.get",
            Mock(
                return_value=Mock(
                    status_code=status_code,
                    ok=status_code < 400,
                    json=lambda: json,
                )
            ),
        )

    @pytest.mark.parametrize("test_case", ["valid", "invalid_scope", "invalid"])
    def test_auth_login_token(self, monkeypatch, cli_fs_runner, test_case):
        """
        GIVEN an API token, valid or not
        WHEN the auth login command is called with --method=token
        THEN the validity of the token should be checked, and if valid, the user should be logged in
        """
        token = "mysupertoken"
        instance = "https://dashboard.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={instance}"]

        if test_case == "valid":
            self.mock_autho_login_request(monkeypatch, 200, self.VALID_TOKEN_PAYLOAD)
        elif test_case == "invalid_scope":
            self.mock_autho_login_request(
                monkeypatch, 200, self.VALID_TOKEN_INVALID_SCOPE_PAYLOAD
            )
        elif test_case == "invalid":
            self.mock_autho_login_request(monkeypatch, 401, self.INVALID_TOKEN_PAYLOAD)

        check_instance_has_enabled_flow_mock = Mock()
        monkeypatch.setattr(
            "ggshield.cmd.auth.login.check_instance_has_enabled_flow",
            check_instance_has_enabled_flow_mock,
        )

        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        if test_case == "valid":
            assert_invoke_ok(result)
            assert instance in config_instance_urls
            assert config.auth_config.get_instance(instance).account.token == token
        else:
            assert result.exit_code != 0
            if test_case == "invalid_scope":
                assert "This token does not have the scan scope." in result.output
            else:
                assert "Authentication failed with token." in result.output
            assert instance not in config_instance_urls

        check_instance_has_enabled_flow_mock.assert_not_called()

    def test_auth_login_token_default_instance(self, monkeypatch, cli_fs_runner):
        """
        GIVEN a valid API token
        WHEN the auth login command is called without --instance with method token
        THEN the authentication is made against the default instance
        """
        config = Config()
        assert len(config.auth_config.instances) == 0

        self.mock_autho_login_request(monkeypatch, 200, self.VALID_TOKEN_PAYLOAD)

        cmd = ["auth", "login", "--method=token"]

        token = "mysupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 1
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        assert config.instance_name in config_instance_urls
        assert (
            config.auth_config.get_instance(config.instance_name).account.token == token
        )

    def test_auth_login_token_update_existing_config(self, monkeypatch, cli_fs_runner):
        """
        GIVEN some valid API tokens
        WHEN the auth login command is called with --method=token
        THEN the instance configuration is created if it doesn't exist, or updated otherwise
        """
        monkeypatch.setattr(
            "ggshield.core.client.GGClient.get",
            Mock(return_value=Mock(ok=True, json=lambda: _TOKEN_RESPONSE_PAYLOAD)),
        )

        instance = "https://dashboard.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={instance}"]

        token = "myfirstsupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert config.auth_config.get_instance(instance).account.token == token

        token = "mysecondsupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 1
        assert config.auth_config.get_instance(instance).account.token == token

        second_instance_token = "mythirdsupertoken"
        second_instance = "https://dashboard.other.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={second_instance}"]
        result = cli_fs_runner.invoke(
            cli, cmd, color=False, input=second_instance_token + "\n"
        )

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 2
        assert config.auth_config.get_instance(instance).account.token == token
        assert (
            config.auth_config.get_instance(second_instance).account.token
            == second_instance_token
        )


class TestAuthLoginWeb:
    @pytest.mark.parametrize("instance_url", [None, "https://some_instance.com"])
    def test_existing_token_no_expiry(self, instance_url, cli_fs_runner, monkeypatch):

        self.prepare_mocks(monkeypatch, instance_url=instance_url)
        prepare_config(instance_url=instance_url)

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 0, output
        self._webbrowser_open_mock.assert_not_called()

        self._assert_last_print(
            output, "ggshield is already authenticated without an expiry date"
        )

    @pytest.mark.parametrize("instance_url", [None, "https://some_instance.com"])
    @pytest.mark.parametrize(
        ["month", "day", "str_date"],
        [
            ("01", "31", "January 31st"),
            ("02", "22", "February 22nd"),
            ("03", "13", "March 13th"),
            ("04", "03", "April 3rd"),
            ("05", "04", "May 4th"),
        ],
    )
    def test_existing_non_expired_token(
        self, month, day, str_date, instance_url, cli_fs_runner, monkeypatch
    ):
        dt = datetime.strptime(f"2100-{month}-{day}T00:00:00+0000", DT_FORMAT)

        self.prepare_mocks(monkeypatch, instance_url=instance_url)
        prepare_config(instance_url=instance_url, expiry_date=dt)

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 0, output
        self._webbrowser_open_mock.assert_not_called()

        self._assert_last_print(
            output, f"ggshield is already authenticated until {str_date} 2100"
        )

    def test_no_port_available_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN -
        WHEN initiating the oauth flow
        AND all potential ports to run the local server are occupied
        THEN the auth flow fails with an explanatory message
        """

        self.prepare_mocks(monkeypatch, used_port_count=1000)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 1

        self._webbrowser_open_mock.assert_not_called()
        self._client_post_mock.assert_not_called()
        self._client_get_mock.assert_not_called()
        self._assert_last_print(output, "Error: Could not find unoccupied port.")
        self._assert_config_is_empty()

    @pytest.mark.parametrize(
        ["authorization_code", "is_state_valid"],
        [
            pytest.param(None, True, id="no-auth-code"),
            pytest.param("some_authorization_code", False, id="invalid-state"),
        ],
    )
    def test_invalid_oauth_params_exits_error(
        self, authorization_code, is_state_valid, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN -
        WHEN receiving the oauth flow callback
        AND the callback doesn't include an authorization code
        OR the state included in the url doesn't match the original state
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(
            monkeypatch,
            authorization_code=authorization_code,
            is_state_valid=is_state_valid,
        )
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 1

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()
        self._client_post_mock.assert_not_called()
        self._client_get_mock.assert_not_called()
        self._assert_last_print(
            output,
            "Error: Invalid code or state received from the callback.",
        )
        self._assert_config_is_empty()

    def test_invalid_code_exchange_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN -
        WHEN receiving the oauth flow callback
        AND the call to the server to exchange the authorization code
        against an access token fails
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(monkeypatch, is_exchange_ok=False)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 1

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()

        self._client_post_mock.assert_called_once()
        self._assert_post_payload()
        self._assert_last_print(output, "Error: Cannot create a token.")
        self._client_get_mock.assert_not_called()

    def test_invalid_token_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN a token created via the oauth process
        WHEN the token is invalid
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(monkeypatch, is_token_valid=False)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 1

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()

        self._client_post_mock.assert_called_once()
        self._assert_post_payload()
        self._assert_last_print(output, "Error: The created token is invalid.")
        self._client_get_mock.assert_called_once()

    @pytest.mark.parametrize("token_name", [None, "some token name"])
    @pytest.mark.parametrize("lifetime", [None, 0, 1, 365])
    @pytest.mark.parametrize("used_port_count", [0, 1, 10])
    @pytest.mark.parametrize("existing_expired_token", [False, True])
    @pytest.mark.parametrize("existing_unrelated_token", [False, True])
    def test_valid_process(
        self,
        existing_unrelated_token,
        existing_expired_token,
        used_port_count,
        lifetime,
        token_name,
        cli_fs_runner,
        monkeypatch,
    ):
        self.prepare_mocks(
            monkeypatch,
            token_name=token_name,
            lifetime=lifetime,
            used_port_count=used_port_count,
        )

        if existing_expired_token:
            # save expired in config
            prepare_config(
                expiry_date=datetime.now(tz=timezone.utc).replace(microsecond=0)
                - timedelta(days=2)
            )
        if existing_unrelated_token:
            # add a dummy unrelated confif
            prepare_config(instance_url="http://some-gg-instance.com")

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 0, output

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url(expected_port=29170 + used_port_count)

        self._client_post_mock.assert_called_once()
        self._assert_post_payload()
        self._client_get_mock.assert_called_once()

        if self._lifetime is None:
            str_date = "never"
        else:
            str_date = get_pretty_date(self._get_expiry_date())

        message = (
            "Success! You are now authenticated.\n"
            "The personal access token has been created and stored in your ggshield config.\n\n"
            f"token name: {self._generated_token_name}\n"
            f"token expiration date: {str_date}\n\n"
            'You do not need to run "ggshield auth login" again. Future requests will automatically use the token.\n'
        )

        assert output.endswith(message)

        self._assert_config("mysupertoken")

    def prepare_mocks(
        self,
        monkeypatch,
        token_name=None,
        lifetime=None,
        instance_url=None,
        authorization_code="some_authorization_code",
        used_port_count=0,
        is_state_valid=True,
        is_exchange_ok=True,
        is_token_valid=True,
        sso_url=None,
    ):
        """
        Prepare object and function mocks to emulate HTTP requests
        and server interactions
        """
        token = "mysupertoken"
        config = Config()
        assert len(config.auth_config.instances) == 0

        # original token params
        self._token_name = token_name
        self._lifetime = lifetime
        self._instance_url = instance_url
        self._sso_url = sso_url

        # token name generated if passed as None
        self._generated_token_name = (
            token_name
            if token_name
            else "ggshield token " + datetime.today().strftime("%Y-%m-%d")
        )

        # generate the expected oauth state
        self._state = self._get_oauth_state() if is_state_valid else "invalid_state"
        url_params = {"state": self._state}
        if authorization_code:
            url_params["code"] = authorization_code

        callback_url = "http://localhost:1234/?" + urlparse.urlencode(url_params)

        # emulates incoming request
        monkeypatch.setattr(
            "ggshield.cmd.auth.login.OAuthClient",
            self._get_oauth_client_class(callback_url),
        )

        # Ensure that original wait_for_code method is not called
        self._wait_for_callback_mock = Mock()
        monkeypatch.setattr(
            "ggshield.core.oauth.OAuthClient._wait_for_callback",
            self._wait_for_callback_mock,
        )

        # open browser for the user to login
        self._webbrowser_open_mock = Mock()
        monkeypatch.setattr(
            "ggshield.core.oauth.webbrowser.open_new_tab", self._webbrowser_open_mock
        )

        # avoid starting a server on port 1234
        self._mock_server_class = Mock(
            side_effect=self._get_oserror_side_effect(used_port_count)
        )
        monkeypatch.setattr("ggshield.core.oauth.HTTPServer", self._mock_server_class)

        token_response_payload = {}
        if is_exchange_ok:
            token_response_payload = _TOKEN_RESPONSE_PAYLOAD.copy()
            if lifetime is not None:
                token_response_payload["expire_at"] = self._get_expiry_date()

        # mock api call to exchange the code against a valid access token
        self._client_post_mock = Mock(
            return_value=Mock(
                ok=is_exchange_ok,
                json=lambda: (
                    {"key": token, **token_response_payload} if is_exchange_ok else {}
                ),
            )
        )
        monkeypatch.setattr("ggshield.core.client.Session.post", self._client_post_mock)

        # mock api call to test the access token
        self._client_get_mock = Mock(
            return_value=Mock(
                ok=is_token_valid,
                json=lambda: token_response_payload,
            )
        )
        monkeypatch.setattr("ggshield.core.client.GGClient.get", self._client_get_mock)

        self._check_instance_has_enabled_flow_mock = Mock()
        monkeypatch.setattr(
            "ggshield.cmd.auth.login.check_instance_has_enabled_flow",
            self._check_instance_has_enabled_flow_mock,
        )

        # run cli command

    def run_cmd(self, cli_fs_runner, method="web", expect_check_feature_flag=True):
        """
        Run the auth login method within a virtual cli.
        Make sure the original server method is not called.
        """
        cmd = ["auth", "login", f"--method={method}"]
        if self._token_name is not None:
            cmd.append(f"--token-name={self._token_name}")

        if self._lifetime is not None:
            cmd.append(f"--lifetime={self._lifetime}")

        if self._instance_url is not None:
            cmd.append(f"--instance={self._instance_url}")

        if self._sso_url is not None:
            cmd.append(f"--sso-url={self._sso_url}")

        # run cli command
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)

        # make sure the first call to api is mocked
        if expect_check_feature_flag:
            self._check_instance_has_enabled_flow_mock.assert_called_once()

        # original method should not be called
        self._wait_for_callback_mock.assert_not_called()
        return result.exit_code, result.output

    @staticmethod
    def _assert_config_is_empty():
        """
        assert that the config is empty
        """
        config = Config()
        assert len(config.auth_config.instances) == 0

    @staticmethod
    def _assert_config(token=None):
        """
        assert that the config exists.
        If a token is passed, assert that the token saved in the config is the same
        """
        config = Config()
        assert len(config.auth_config.instances) >= 1
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        assert config.instance_name in config_instance_urls
        if token is not None:
            assert (
                config.auth_config.get_instance(config.instance_name).account.token
                == token
            )

    @staticmethod
    def _get_oserror_side_effect(failure_count=1):
        """
        return a side effect to pass to a mock object
        the n first call will raise an exception
        the call n + 1 will be silent
        """
        return (
            OSError("This port is already in use.") if i < failure_count else None
            for i in range(failure_count + 1)
        )

    @staticmethod
    def _assert_last_print(output: str, expected_str: str):
        """
        assert that the last log output is the same as the one passed in param
        """
        assert output.rsplit("\n", 2)[-2] == expected_str

    def _assert_open_url(
        self, *, host: Optional[str] = None, expected_port: int = 29170
    ):
        """
        assert that the url to be open in the browser has the right static parameters
        also check if the port of the redirect url is the one expected depending on occupied ports
        """
        (url,), kwargs = self._webbrowser_open_mock.call_args_list[0]
        parsed_url = urlparse.urlparse(url)
        url_params = urlparse.parse_qs(parsed_url.query)

        if host is not None:
            assert (
                host == parsed_url.netloc
            ), f"invalid host opened: '{parsed_url.netloc}', expected '{host}'"

        for key, value in _EXPECTED_URL_PARAMS.items():
            assert key in url_params, f"missing url param: {key}"
            assert url_params[key] == value, f"invalid value for param '{key}': {value}"

        assert "redirect_uri" in url_params, "redirect_url not in server post request"
        redirect_uri = url_params["redirect_uri"][0]

        assert redirect_uri.startswith(
            f"http://localhost:{expected_port}"
        ), redirect_uri

    def _assert_post_payload(self):
        """
        assert that the post request is made to the right url
        and include the expected token name
        """

        (url, payload), kwargs = self._client_post_mock.call_args_list[0]
        assert url == "https://api.gitguardian.com/v1/oauth/token"

        request_body = urlparse.parse_qs(payload)

        assert "name" in request_body
        assert request_body["name"][0] == self._generated_token_name

        if self._lifetime is None:
            assert "lifetime" not in request_body
        else:
            assert "lifetime" in request_body
            assert request_body["lifetime"][0] == str(self._lifetime)

    def _get_expiry_date(self) -> Optional[datetime]:
        """
        Get the token expiry date based on the given lifetime
        the date is offset to jan 1st 2100 to make testing easier
        """
        if self._lifetime is not None:
            return datetime.strptime("2100-01-01T00:00:00+0000", DT_FORMAT) + timedelta(
                days=self._lifetime
            )
        return None

    def _get_oauth_state(self):
        return urlparse.quote(
            json.dumps(
                {"token_name": self._generated_token_name, "lifetime": self._lifetime}
            )
        )

    @staticmethod
    def _get_oauth_client_class(callback_url):
        """
        generate a fake OAuth class which overrides the wait for callback function
        instead of waiting for a callback, emulate a callback with the given url
        """

        class FakeOAuthClient(OAuthClient):
            def _wait_for_callback(self, *args, **kwargs):
                try:
                    self.process_callback(callback_url)
                except OAuthError as e:
                    raise ClickException(e.message)

        return FakeOAuthClient

    @pytest.mark.parametrize(
        ["version", "preference_enabled", "status_code", "expected_error"],
        [
            ["2022.04.0", True, 200, VERSION_TOO_LOW_MESSAGE],
            ["v1.0.0", None, 404, VERSION_TOO_LOW_MESSAGE],
            ["v1.0.0", False, 200, DISABLED_FLOW_MESSAGE],
            ["v1.0.0", True, 200, None],
            [
                "v1.0.0",
                None,
                200,
                None,
            ],  # Removing the preference is like making it on by default
        ],
    )
    def test_assert_flow_enabled(
        self,
        monkeypatch,
        version,
        preference_enabled,
        status_code,
        expected_error,
    ):
        """
        GIVEN -
        WHEN checking the availability of the ggshield auth flow web method on a dashboard of various
        various, with or without the flow enabled
        THEN it succeeds if the version is high enough, and the preference is enabled
        """

        def client_get_mock(self_, url, **kwargs):
            if url.endswith("/v1/metadata"):
                return Mock(
                    ok=status_code < 400,
                    status_code=status_code,
                    json=lambda: {
                        "version": version,
                        "preferences": {
                            "public_api__ggshield_auth_flow_enabled": preference_enabled
                        }
                        if preference_enabled is not None
                        else {},
                    },
                )
            raise NotImplementedError

        monkeypatch.setattr("ggshield.core.client.Session.get", client_get_mock)

        if expected_error:
            with pytest.raises(ClickException, match=expected_error):
                check_instance_has_enabled_flow(Config())
        else:
            check_instance_has_enabled_flow(Config())

    @pytest.mark.parametrize(
        ["method", "instance_url", "sso_url", "expected_error"],
        [
            [
                "web",
                "https://dashboard.gitguardian.com",
                "https://onprem.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "Error: instance and SSO URL params do not match",
            ],
            [
                "web",
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com",
                "Error: Invalid value for sso-url: Please provide a valid SSO URL.",
            ],
            [
                "token",
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "Error: Invalid value for sso-url: --sso-url is reserved for the web login method.",
            ],
        ],
    )
    def test_bad_sso_url(
        self, method, instance_url, sso_url, expected_error, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN an invalid SSO URL, or that do not match the declared instance
        WHEN running the login command
        THEN it fails
        """
        self.prepare_mocks(monkeypatch, instance_url=instance_url, sso_url=sso_url)
        prepare_config(instance_url=instance_url)

        exit_code, output = self.run_cmd(
            cli_fs_runner, method=method, expect_check_feature_flag=False
        )
        assert exit_code > 0, output
        self._webbrowser_open_mock.assert_not_called()
        self._assert_last_print(output, expected_error)

    @pytest.mark.parametrize(
        ["instance_url", "sso_url", "expected_web_host"],
        [
            [
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "dashboard.gitguardian.com",
            ],
            [
                "https://some-gg-instance.com",
                "https://some-gg-instance.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "some-gg-instance.com",
            ],
            [
                None,
                "https://custom.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "custom.gitguardian.com",
            ],
            [
                None,
                None,
                "dashboard.gitguardian.com",
            ],
        ],
    )
    def test_sso_url(
        self, instance_url, sso_url, expected_web_host, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN -
        WHEN calling the login command on the web method with a valid SSO URL, or no SSO URL
        THEN if no SSO URL is passed, the default instance is used
        THEN if the SSO URL is compatible with the instance, the instance is used
        THEN if no instance is declared, the instance of the SSO URL is used
        THEN the correct instance URL is checked to see if the auth flow  is enabled
        """
        self.prepare_mocks(monkeypatch, instance_url=instance_url, sso_url=sso_url)
        config_dashboard_url_checked = None

        def setter(config):
            nonlocal config_dashboard_url_checked
            config_dashboard_url_checked = config.dashboard_url

        # store the config dashboard URL at the time the check is done because the config
        # object is mutable
        self._check_instance_has_enabled_flow_mock.side_effect = setter

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == 0, output
        self._webbrowser_open_mock.assert_called()
        self._assert_open_url(host=expected_web_host)
        self._check_instance_has_enabled_flow_mock.assert_called_once()
        assert (
            urlparse.urlparse(config_dashboard_url_checked).netloc == expected_web_host
        )


class TestLoginUtils:
    @pytest.mark.parametrize(
        ["url", "expected_error"],
        [
            ("http://localhost:3455", None),
            ("http://localhost:3455?", None),
            ("http://localhost:3455?auth=ggshield", None),
            ("http://localhost:3455?error=some+error", "some error"),
            ("http://localhost/?error=some+error", "some error"),
            ("http://localhost:3455/?auth=ggshield&error=some+error", "some error"),
        ],
    )
    def test_get_error_url_param(self, url, expected_error):
        """
        GIVEN a url
        WHEN calling get_error_param
        THEN it returns the value of the 'error' parameter if it exists else None
        """
        error = get_error_param(urlparse.urlparse(url))
        assert error == expected_error

    @pytest.mark.parametrize(
        ["error_code", "expected_message"],
        [
            (
                "too_many_tokens",
                (
                    "Maximum number of personal access tokens reached. "
                    "Could not provision a new personal access token.\n"
                    "Go to your workspace to manage your tokens: "
                    "https://dashboard.gitguardian.com/api/personal-access-tokens"
                ),
            ),
            (
                "invalid_saml",
                "The given SSO URL is invalid.",
            ),
            (
                "invalid_error_code",
                "An unknown server error has occurred (error code: invalid_error_code).",
            ),
        ],
    )
    def test_get_error_message(self, error_code, expected_message):
        """
        GIVEN an OAuthClient instance and an error code
        WHEN calling OAuthClient.get_server_error with the error code
        THEN it should return the corresponding human readable message with formated urls
        """
        oauth_client = OAuthClient(Config(), "https://dashboard.gitguardian.com")
        error_message = oauth_client.get_server_error_message(error_code)
        assert error_message == expected_message



import os


# max files size to create a tar from
MAX_TAR_CONTENT_SIZE = 30 * 1024 * 1024

CPU_COUNT = os.cpu_count() or 1


def _get_max_workers() -> int:
    # We add a max value to avoid silently consuming all threads on powerful machines
    return int(os.getenv("GG_MAX_WORKERS", "32"))


MAX_WORKERS = min(CPU_COUNT, _get_max_workers())
CACHE_FILENAME = "./.cache_ggshield"
GLOBAL_CONFIG_FILENAMES = [".gitguardian", ".gitguardian.yml", ".gitguardian.yaml"]
LOCAL_CONFIG_PATHS = ["./.gitguardian", "./.gitguardian.yml", "./.gitguardian.yaml"]
DEFAULT_LOCAL_CONFIG_PATH = "./.gitguardian.yaml"
DEFAULT_DASHBOARD_URL = "https://dashboard.gitguardian.com"
AUTH_CONFIG_FILENAME = "auth_config.yaml"
ON_PREMISE_API_URL_PATH_PREFIX = "/exposed"



import json
from unittest import mock

import pytest
from pytest_voluptuous import Partial, S
from voluptuous.validators import All, Invalid, Match, Range

from ggshield.cmd.main import cli
from tests.conftest import assert_invoke_ok, my_vcr


def test_quota(cli_fs_runner):
    with my_vcr.use_cassette("quota"):
        cmd = ["quota", "--json"]
        result = cli_fs_runner.invoke(cli, cmd, color=False)
        assert_invoke_ok(result)

        def quota_values_must_match(output):
            if output["count"] + output["remaining"] != output["limit"]:
                raise Invalid("API calls count and remaining must sum to limit.")

        assert S(
            All(
                Partial(  # Partial validation because of the "since" key
                    {
                        "count": All(int, Range(min=0)),
                        "limit": All(int, Range(min=0)),
                        "remaining": All(int, Range(min=0)),
                    }
                ),
                quota_values_must_match,
            )
        ) == json.loads(result.output)


def test_api_status(cli_fs_runner):
    with my_vcr.use_cassette("test_health_check"):
        cmd = ["api-status", "--json"]
        result = cli_fs_runner.invoke(cli, cmd, color=False)
        assert_invoke_ok(result)

        assert S(
            All(
                Partial(  # Partial validation because of the "since" key
                    {
                        "detail": "Valid API key.",
                        "status_code": 200,
                        "app_version": Match(r"v\d\.\d{1,3}\.\d{1,2}(-rc\.\d)?"),
                        "secrets_engine_version": Match(r"\d\.\d{1,3}\.\d"),
                    }
                ),
            )
        ) == json.loads(result.output)


@pytest.mark.parametrize("verify", [True, False])
def test_ssl_verify(cli_fs_runner, verify):
    cmd = ["api-status"] if verify else ["--allow-self-signed", "api-status"]

    with mock.patch("ggshield.core.client.IaCGGClient") as client_mock:
        cli_fs_runner.invoke(cli, cmd)
        _, kwargs = client_mock.call_args
        assert kwargs["session"].verify == verify



import os
import subprocess
import sys
from typing import Optional

import click

from ggshield.core.git_shell import GIT_PATH, check_git_dir, check_git_installed


@click.command(context_settings={"ignore_unknown_options": True})
@click.option(
    "--mode",
    "-m",
    type=click.Choice(["local", "global"]),
    help="Hook installation mode",
    required=True,
)
@click.option(
    "--hook-type",
    "-t",
    type=click.Choice(["pre-commit", "pre-push"]),
    help="Type of hook to install",
    default="pre-commit",
)
@click.option("--force", "-f", is_flag=True, help="Force override")
@click.option("--append", "-a", is_flag=True, help="Append to existing script")
def install_cmd(mode: str, hook_type: str, force: bool, append: bool) -> int:
    """Install a pre-commit or pre-push git hook (local or global)."""
    return_code = (
        install_global(hook_type=hook_type, force=force, append=append)
        if mode == "global"
        else install_local(hook_type=hook_type, force=force, append=append)
    )
    sys.exit(return_code)


def install_global(hook_type: str, force: bool, append: bool) -> int:
    """Global pre-commit/pre-push hook installation."""
    check_git_installed()
    hook_dir_path = get_global_hook_dir_path()

    if not hook_dir_path:
        hook_dir_path = os.path.expanduser("~/.git/hooks")
        subprocess.run(
            [GIT_PATH, "config", "--global", "core.hooksPath", hook_dir_path]
        )

    return create_hook(
        hook_dir_path=hook_dir_path,
        force=force,
        local_hook_support=True,
        hook_type=hook_type,
        append=append,
    )


def get_global_hook_dir_path() -> Optional[str]:
    """Return the default hooks path (if it exists)."""
    with subprocess.Popen(
        [GIT_PATH, "config", "--global", "--get", "core.hooksPath"],
        stdout=subprocess.PIPE,
    ) as process:
        if process.returncode:
            return None

        return os.path.expanduser(
            click.format_filename(process.communicate()[0].decode("utf-8")).strip()
        )


def install_local(hook_type: str, force: bool, append: bool) -> int:
    """Local pre-commit/pre-push hook installation."""
    check_git_dir()
    return create_hook(
        hook_dir_path=".git/hooks",
        force=force,
        local_hook_support=False,
        hook_type=hook_type,
        append=append,
    )


def create_hook(
    hook_dir_path: str,
    force: bool,
    local_hook_support: bool,
    hook_type: str,
    append: bool,
) -> int:
    """Create hook directory (if needed) and pre-commit/pre-push file."""
    os.makedirs(hook_dir_path, exist_ok=True)
    hook_path = f"{hook_dir_path}/{hook_type}"

    if os.path.isdir(hook_path):
        raise click.ClickException(f"{hook_path} is a directory.")

    if os.path.isfile(hook_path) and not (force or append):
        raise click.ClickException(
            f"{hook_path} already exists."
            " Use --force to override or --append to add to current script"
        )

    local_hook_str = ""
    if local_hook_support:
        local_hook_str = f"""
if [[ -f .git/hooks/{hook_type} ]]; then
    if ! .git/hooks/{hook_type} $@; then
        echo 'Local {hook_type} hook failed, please see output above'
        exit 1
    fi
fi
"""

    mode = "a" if os.path.exists(hook_path) and append else "w"

    with open(hook_path, mode) as f:
        if mode == "w":
            f.write("#!/bin/bash\n")

        f.write(f"\n{local_hook_str}\nggshield secret scan {hook_type}\n")
        os.chmod(hook_path, 0o700)

    click.echo(
        f"{hook_type} successfully added in"
        f" {click.style(hook_path, fg='yellow', bold=True)}"
    )

    return 0



from typing import Dict, Optional, Union, cast

import click
import requests
import urllib3
from pygitguardian import GGClient
from pygitguardian.client import is_ok
from pygitguardian.models import Detail
from requests import Response, Session

from ..iac.models import IaCScanResult, IaCScanResultSchema
from ..iac.models.iac_scan_parameters import IaCScanParameters, IaCScanParametersSchema
from .config import Config
from .config.errors import UnknownInstanceError
from .constants import DEFAULT_DASHBOARD_URL


def load_detail(resp: Response) -> Detail:
    """
    load_detail loads a Detail from a response
    be it JSON or html.

    :param resp: API response
    :type resp: Response
    :return: detail object of response
    :rtype: Detail
    """
    if resp.headers["content-type"] == "application/json":
        data = resp.json()
        if "detail" not in data:
            data = {"detail": str(data)}
    else:
        data = {"detail": resp.text}

    return cast(Detail, Detail.SCHEMA.load(data))


def create_client_from_config(config: Config) -> GGClient:
    """
    Create a GGClient using parameters from Config.
    """
    try:
        api_key = config.api_key
        api_url = config.api_url
    except UnknownInstanceError as e:
        if e.instance == DEFAULT_DASHBOARD_URL:
            # This can happen when the user first tries the app and has not gone through
            # the authentication procedure yet. In this case, replace the error message
            # complaining about an unknown instance with a more user-friendly one.
            raise click.ClickException("GitGuardian API key is needed.")
        else:
            raise

    return create_client(api_key, api_url, allow_self_signed=config.allow_self_signed)


def create_client(
    api_key: str, api_url: str, *, allow_self_signed: bool = False
) -> GGClient:
    """
    Implementation of create_client_from_config(). Exposed as a function for specific
    cases such as needing a GGClient instance while defining the config account.
    """
    session = create_session(allow_self_signed=allow_self_signed)
    try:
        return IaCGGClient(
            api_key=api_key,
            base_uri=api_url,
            user_agent="ggshield",
            timeout=60,
            session=session,
        )
    except ValueError as e:
        # Can be raised by pygitguardian
        raise click.ClickException(f"Failed to create API client. {e}")


def create_session(allow_self_signed: bool = False) -> Session:
    session = Session()
    if allow_self_signed:
        urllib3.disable_warnings()
        session.verify = False
    return session


class IaCGGClient(GGClient):
    def directory_scan(
        self,
        directory: bytes,
        scan_parameters: IaCScanParameters,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Union[Detail, IaCScanResult]:

        result: Union[Detail, IaCScanResult]
        try:
            resp = self.request(
                "post",
                endpoint="iac_scan",
                extra_headers=extra_headers,
                files={
                    "directory": directory,
                },
                data={
                    "scan_parameters": IaCScanParametersSchema().dumps(scan_parameters),
                },
            )
        except requests.exceptions.ReadTimeout:
            result = Detail("The request timed out.")
            result.status_code = 504
        else:
            if is_ok(resp):
                result = IaCScanResultSchema().load(resp.json())
            else:
                result = load_detail(resp)

            result.status_code = resp.status_code

        return result



from typing import List

from ggshield.core.text_utils import LINE_DISPLAY, Line


def get_padding(lines: List[Line]) -> int:
    """Return the number of digit of the maximum line number."""
    # value can be None
    return max(len(str(lines[-1].pre_index or 0)), len(str(lines[-1].post_index or 0)))


def get_offset(padding: int, is_patch: bool = False) -> int:
    """Return the offset due to the line display."""
    if is_patch:
        return len(LINE_DISPLAY["patch"].format("0" * padding, "0" * padding))

    return len(LINE_DISPLAY["file"].format("0" * padding))



import click
from requests.exceptions import ConnectionError

from ggshield.core.client import create_client
from ggshield.core.config import Config
from ggshield.core.utils import dashboard_to_api_url

from .utils import CONNECTION_ERROR_MESSAGE


@click.command()
@click.option(
    "--instance",
    required=False,
    type=str,
    help="URL of the instance to logout from.",
)
@click.option(
    "--revoke/--no-revoke",
    is_flag=True,
    default=True,
    help="Whether the token should be revoked on logout before being removed from the configuration.",
)
@click.option("--all", "all_", is_flag=True, help="Iterate over every saved tokens.")
@click.pass_context
def logout_cmd(ctx: click.Context, instance: str, revoke: bool, all_: bool) -> int:
    """
    Remove authentication for a GitGuardian instance.
    A successful logout results in the deletion of personal access token stored in the configuration.
    By default, the token will be revoked unless `--no-revoke` option is specified.

    If not specified, ggshield will logout from the default instance.
    The `--all` option can be used if you want to logout from all your GitGuardian instances.
    """
    config = ctx.obj["config"]

    if all_:
        for _instance in config.auth_config.instances:
            logout(config, _instance.url, revoke=revoke)
    else:
        if not instance:
            instance = config.instance_name
        logout(config, instance, revoke=revoke)
    return 0


def logout(config: Config, instance_url: str, revoke: bool) -> None:
    check_account_config_exists(config, instance_url)
    if revoke:
        revoke_token(config, instance_url)
    delete_account_config(config, instance_url)

    click.echo(
        f"Successfully logged out for instance {instance_url}\n\n"
        f"Your personal access token has been{' revoked and' if revoke else ''} "
        "removed from your configuration."
    )


def check_account_config_exists(config: Config, instance_url: str) -> None:
    instance = config.auth_config.get_instance(instance_url)
    if instance.account is None:
        raise click.ClickException(
            f"No token found for instance {instance_url}\n"
            "First try to login by running:\n"
            "  ggshield auth login"
        )


def revoke_token(config: Config, instance_url: str) -> None:

    instance = config.auth_config.get_instance(instance_url)

    assert instance.account is not None
    token = instance.account.token

    client = create_client(
        token,
        dashboard_to_api_url(instance_url),
        allow_self_signed=config.allow_self_signed,
    )
    try:
        response = client.post(endpoint="token/revoke")
    except ConnectionError:
        raise click.ClickException(CONNECTION_ERROR_MESSAGE)

    if response.status_code != 204:
        raise click.ClickException(
            "Could not perform the logout command because your token is already revoked or invalid.\n"
            "Please try with the following command:\n"
            "  ggshield auth logout --no-revoke"
        )


def delete_account_config(config: Config, instance: str) -> None:
    instance_config = config.auth_config.get_instance(instance)
    account_config = instance_config.account

    assert account_config is not None

    instance_config.account = None
    config.auth_config.set_instance(instance_config)
    config.auth_config.save()



from io import StringIO
from typing import ClassVar, List, cast

import click
from pygitguardian.models import Match

from ggshield.core.filter import censor_content, leak_dictionary_by_ignore_sha
from ggshield.core.text_utils import Line
from ggshield.core.utils import Filemode, find_match_indices, get_lines_from_content
from ggshield.output.output_handler import OutputHandler
from ggshield.scan import Result, Results, ScanCollection

from .message import (
    file_info,
    flatten_policy_breaks_by_line,
    leak_message_located,
    no_leak_message,
    policy_break_header,
    secrets_engine_version,
)
from .utils import get_offset, get_padding


class TextOutputHandler(OutputHandler):
    nb_lines: ClassVar[int] = 3

    def _process_scan_impl(self, scan: ScanCollection, top: bool = True) -> str:
        scan_buf = StringIO()
        if scan.optional_header and (scan.has_results or self.verbose):
            scan_buf.write(scan.optional_header)

        if top and (scan.has_results or self.verbose):
            scan_buf.write(secrets_engine_version())

        if scan.has_results:
            for result in cast(Results, scan.results).results:
                scan_buf.write(self.process_result(result))
        else:
            has_results = False
            if scan.scans:
                has_results = any(x.has_results for x in scan.scans)

            if top and not has_results:
                scan_buf.write(no_leak_message())

        if scan.scans:
            for sub_scan in scan.scans:
                inner_scan_str = self._process_scan_impl(sub_scan, top=False)
                scan_buf.write(inner_scan_str)

        return scan_buf.getvalue()

    def process_result(self, result: Result) -> str:
        """
        Build readable message on the found incidents.

        :param result: The result from scanning API
        :param nb_lines: The number of lines to display before and after a secret in the
        patch
        :param show_secrets: Option to show secrets value
        :return: The formatted message to display
        """
        result_buf = StringIO()
        policy_breaks = result.scan.policy_breaks
        is_patch = result.filemode != Filemode.FILE
        sha_dict = leak_dictionary_by_ignore_sha(policy_breaks)

        if self.show_secrets:
            content = result.content
        else:
            content = censor_content(result.content, result.scan.policy_breaks)

        lines = get_lines_from_content(content, result.filemode, is_patch)
        padding = get_padding(lines)
        offset = get_offset(padding, is_patch)

        if len(lines) == 0:
            raise click.ClickException("Parsing of scan result failed.")

        result_buf.write(file_info(result.filename, len(sha_dict)))

        for issue_n, (ignore_sha, policy_breaks) in enumerate(sha_dict.items(), 1):
            result_buf.write(policy_break_header(issue_n, policy_breaks, ignore_sha))
            for policy_break in policy_breaks:
                policy_break.matches = TextOutputHandler.make_matches(
                    policy_break.matches, lines, is_patch
                )

            if policy_breaks[0].policy == "Secrets detection":
                result_buf.write(
                    leak_message_located(
                        flatten_policy_breaks_by_line(policy_breaks),
                        lines,
                        padding,
                        offset,
                        self.nb_lines,
                        clip_long_lines=not self.verbose,
                    )
                )

        return result_buf.getvalue()

    @staticmethod
    def make_matches(
        matches: List[Match], lines: List[Line], is_patch: bool
    ) -> List[Match]:
        res = []
        for match in matches:
            if match.index_start is None or match.index_end is None:
                res.append(match)
                continue
            indices = find_match_indices(match, lines, is_patch)
            res.append(
                Match(
                    match=match.match,
                    match_type=match.match_type,
                    index_start=indices.index_start,
                    index_end=indices.index_end,
                    line_start=indices.line_index_start,
                    line_end=indices.line_index_end,
                )
            )
        return res



import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import click
import marshmallow_dataclass
from marshmallow import ValidationError

from ggshield.core.config.errors import ParseError, format_validation_error
from ggshield.core.config.utils import (
    get_global_path,
    load_yaml,
    remove_common_dict_items,
    save_yaml,
    update_from_other_instance,
)
from ggshield.core.constants import (
    DEFAULT_LOCAL_CONFIG_PATH,
    GLOBAL_CONFIG_FILENAMES,
    LOCAL_CONFIG_PATHS,
)
from ggshield.core.types import FilteredConfig, IgnoredMatch, IgnoredMatchSchema
from ggshield.core.utils import api_to_dashboard_url
from ggshield.iac.utils import POLICY_ID_PATTERN, validate_policy_id


logger = logging.getLogger(__name__)
CURRENT_CONFIG_VERSION = 2


@marshmallow_dataclass.dataclass
class SecretConfig(FilteredConfig):
    """
    Holds all user-defined secret-specific settings
    """

    show_secrets: bool = False
    ignored_detectors: Set[str] = field(default_factory=set)
    ignored_matches: List[IgnoredMatch] = field(default_factory=list)
    ignored_paths: Set[str] = field(default_factory=set)

    def add_ignored_match(self, secret: IgnoredMatch) -> None:
        """
        Add secret to ignored_matches.
        """
        for match in self.ignored_matches:
            if match.match == secret.match:
                # take the opportunity to name the ignored match
                if not match.name:
                    match.name = secret.name
                return
        self.ignored_matches.append(secret)


def validate_policy_ids(values: Iterable[str]) -> None:
    invalid_excluded_policies = [
        policy_id for policy_id in values if not validate_policy_id(policy_id)
    ]
    if len(invalid_excluded_policies) > 0:
        raise ValidationError(
            f"The policies {invalid_excluded_policies} do not match the pattern '{POLICY_ID_PATTERN.pattern}'"
        )


@marshmallow_dataclass.dataclass
class IaCConfig(FilteredConfig):
    """
    Holds the iac config as defined .gitguardian.yaml files
    (local and global).
    """

    ignored_paths: Set[str] = field(default_factory=set)
    ignored_policies: Set[str] = field(
        default_factory=set, metadata={"validate": validate_policy_ids}
    )
    minimum_severity: str = "LOW"


@marshmallow_dataclass.dataclass
class UserConfig(FilteredConfig):
    """
    Holds all ggshield settings defined by the user in the .gitguardian.yaml files
    (local and global).
    """

    iac: IaCConfig = field(default_factory=IaCConfig)
    instance: Optional[str] = None
    exit_zero: bool = False
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    secret: SecretConfig = field(default_factory=SecretConfig)
    debug: bool = False

    # If we hit any deprecated syntax when loading a configuration file, we do not
    # display them directly, otherwise the messages would also be shown when running
    # `ggshield config migrate`, which would be odd.
    # Instead, we store them in this list and a result_callback() function displays
    # them when we quit. When `config migrate` runs, it clears this list, so nothing
    # gets displayed.
    deprecation_messages: List[str] = field(default_factory=list)

    def save(self, config_path: str) -> None:
        """
        Save config to config_path
        """
        schema = UserConfigSchema(exclude=("deprecation_messages",))
        dct = schema.dump(self)
        default_dct = schema.dump(schema.load({}))

        dct = remove_common_dict_items(dct, default_dct)

        dct["version"] = CURRENT_CONFIG_VERSION

        save_yaml(dct, config_path)

    @classmethod
    def load(cls, config_path: Optional[str] = None) -> Tuple["UserConfig", str]:
        """
        Load the various user configs files to create a UserConfig object:
        - global user configuration file (in the home)
        - local user configuration file (in the repository)

        Returns a UserConfig instance, and the path where updates should be saved
        """
        user_config = UserConfig()
        if config_path:
            logger.debug("Loading custom config from %s", config_path)
            user_config._update_from_file(config_path)
            return user_config, config_path

        for global_config_filename in GLOBAL_CONFIG_FILENAMES:
            global_config_path = get_global_path(global_config_filename)
            if os.path.exists(global_config_path):
                user_config._update_from_file(global_config_path)
                logger.debug("Loaded global config from %s", global_config_path)
                break
        else:
            logger.debug("No global config")

        for local_config_path in LOCAL_CONFIG_PATHS:
            if os.path.exists(local_config_path):
                user_config._update_from_file(local_config_path)
                config_path = local_config_path
                logger.debug("Loaded local config from %s", local_config_path)
                break
        else:
            logger.debug("No local config")

        if config_path is None:
            config_path = DEFAULT_LOCAL_CONFIG_PATH
        return user_config, config_path

    def _update_from_file(self, config_path: str) -> None:
        data = load_yaml(config_path) or {"version": CURRENT_CONFIG_VERSION}
        config_version = data.pop("version", 1)

        try:
            if config_version == 2:
                obj = UserConfigSchema().load(data)
            elif config_version == 1:
                self.deprecation_messages.append(
                    f"{config_path} uses a deprecated configuration file format."
                    " Run `ggshield config migrate` to migrate it to the latest version."
                )
                obj = UserV1Config.load_v1(data)
            else:
                raise click.ClickException(
                    f"Don't know how to load config version {config_version}"
                )
        except ValidationError as exc:
            message = format_validation_error(exc)
            raise ParseError(f"Error in {config_path}:\n{message}") from exc

        update_from_other_instance(self, obj)


UserConfigSchema = marshmallow_dataclass.class_schema(UserConfig)


@dataclass
class UserV1Config:
    """
    Can load a v1 .gitguardian.yaml file
    """

    instance: Optional[str] = None
    all_policies: bool = False
    exit_zero: bool = False
    matches_ignore: List[Dict[str, Optional[str]]] = field(default_factory=list)
    paths_ignore: Set[str] = field(default_factory=set)
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    ignore_default_excludes: bool = False
    show_secrets: bool = False
    banlisted_detectors: Set[str] = field(default_factory=set)

    @staticmethod
    def load_v1(data: Dict[str, Any]) -> UserConfig:
        """
        Takes a dict representing a v1 .gitguardian.yaml and returns a v2 config object
        """
        # If data contains the old "api-url" key, turn it into an "instance" key,
        # but only if there is no "instance" key
        try:
            api_url = data.pop("api_url")
        except KeyError:
            pass
        else:
            if "instance" not in data:
                data["instance"] = api_to_dashboard_url(api_url, warn=True)

        UserV1Config.matches_ignore_to_dict(data)

        v1config = UserV1ConfigSchema().load(data)

        deprecation_messages = []

        if v1config.all_policies:
            deprecation_messages.append(
                "The `all_policies` option has been deprecated and is now ignored."
            )

        if v1config.ignore_default_excludes:
            deprecation_messages.append(
                "The `ignore_default_exclude` option has been deprecated and is now ignored."
            )

        ignored_match_schema = IgnoredMatchSchema()
        ignored_matches = [
            ignored_match_schema.load(secret) for secret in v1config.matches_ignore
        ]
        secret = SecretConfig(
            show_secrets=v1config.show_secrets,
            ignored_detectors=v1config.banlisted_detectors,
            ignored_matches=ignored_matches,
            ignored_paths=v1config.paths_ignore,
        )

        return UserConfig(
            instance=v1config.instance,
            exit_zero=v1config.exit_zero,
            verbose=v1config.verbose,
            allow_self_signed=v1config.allow_self_signed,
            max_commits_for_hook=v1config.max_commits_for_hook,
            secret=secret,
            deprecation_messages=deprecation_messages,
        )

    @staticmethod
    def matches_ignore_to_dict(data: Dict[str, Any]) -> None:
        """
        v1 config format allowed to use just a hash of the secret for matches_ignore
        field v2 does not. This function converts the hash-only matches.
        """
        matches_ignore = data.get("matches_ignore")
        if not matches_ignore:
            return

        for idx, match in enumerate(matches_ignore):
            if isinstance(match, str):
                matches_ignore[idx] = {"name": "", "match": match}


UserV1ConfigSchema = marshmallow_dataclass.class_schema(UserV1Config)



import re
from typing import Optional, Tuple

import click

from ggshield.cmd.auth.utils import check_instance_has_enabled_flow
from ggshield.core.client import create_client
from ggshield.core.config import AccountConfig, Config
from ggshield.core.oauth import OAuthClient
from ggshield.core.utils import clean_url


def validate_login_path(
    instance: Optional[str], sso_url: Optional[str]
) -> Tuple[Optional[str], Optional[str]]:
    """
    Validate that the SSO URL and the instance refer to the same instance if they are both defined,
    that the SSO URL has a correct format
    and return the couple (instance, login_path) that will be used to redirect to the login page
    """
    if sso_url is None:
        return instance, None
    sso_parsed_url = clean_url(sso_url)
    if (
        not sso_parsed_url.scheme
        or not sso_parsed_url.netloc
        or not re.match(
            r"/auth/sso/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
            sso_parsed_url.path,
        )
    ):
        raise click.BadParameter(
            "Please provide a valid SSO URL.",
            param_hint="sso-url",
        )
    sso_instance = f"{sso_parsed_url.scheme}://{sso_parsed_url.netloc}"
    sso_login_path = sso_parsed_url.path

    if instance is None:
        return sso_instance, sso_login_path

    config_parsed_url = clean_url(instance)
    if (
        config_parsed_url.scheme != sso_parsed_url.scheme
        or config_parsed_url.netloc != sso_parsed_url.netloc
    ):
        raise click.ClickException("instance and SSO URL params do not match")
    return instance, sso_login_path


@click.command()
@click.option(
    "--method",
    required=False,
    default="web",
    type=click.Choice(["token", "web"]),
    help="Authentication method.",
)
@click.option(
    "--instance",
    required=False,
    type=str,
    help="URL of the instance to authenticate against.",
)
@click.option(
    "--sso-url",
    required=False,
    type=str,
    help="URL of your SSO login page to force the authentication flow through your workspace SSO.",
)
@click.option(
    "--token-name",
    required=False,
    type=str,
    help="Name of new token.",
)
@click.option(
    "--lifetime",
    required=False,
    type=click.IntRange(0),
    default=None,
    help="Number of days before the token expires. 0 means the token never expires.",
)
@click.pass_context
def login_cmd(
    ctx: click.Context,
    method: str,
    instance: Optional[str],
    token_name: Optional[str],
    lifetime: Optional[int],
    sso_url: Optional[str],
) -> int:
    """
    Authenticate with a GitGuardian workspace.
    A successful authentication results in a personal access token.
    This token is stored in your configuration and used to authenticate your future requests.

    The default authentication method is "web".
    ggshield launches a web browser to authenticate you to your GitGuardian workspace,
    then automatically generates a token on your behalf.

    Alternatively, you can use `--method token` to authenticate using an already existing token.
    The minimum required scope for the token is `scan`.
    """
    config: Config = ctx.obj["config"]

    if sso_url is not None and method != "web":
        raise click.BadParameter(
            "--sso-url is reserved for the web login method.", param_hint="sso-url"
        )

    if method == "token":

        if instance:
            config.set_cmdline_instance_name(instance)
        instance = config.instance_name
        # Override instance to make sure we get a normalized instance name
        instance_config = config.auth_config.get_or_create_instance(
            instance_name=instance
        )
        token = click.prompt("Enter your GitGuardian API token", hide_input=True)
        if not token:
            raise click.ClickException("No API token was provided.")

        # enforce using the token (and not use config default)
        client = create_client(api_key=token, api_url=config.api_url)
        response = client.get(endpoint="token")
        if not response.ok:
            raise click.ClickException("Authentication failed with token.")

        api_token_data = response.json()
        scopes = api_token_data["scope"]
        if "scan" not in scopes:
            raise click.ClickException("This token does not have the scan scope.")

        account_config = AccountConfig(
            workspace_id=api_token_data.get("account_id"),
            token=token,
            expire_at=api_token_data.get("expire_at"),
            token_name=api_token_data.get("name", ""),
            type=api_token_data.get("type", ""),
        )

        instance_config.account = account_config
        config.auth_config.save()
        click.echo("Authentication was successful.")
        return 0

    if method == "web":
        instance, login_path = validate_login_path(instance=instance, sso_url=sso_url)
        if instance:
            config.set_cmdline_instance_name(instance)
        defined_instance = config.instance_name
        # Override instance to make sure we get a normalized instance name

        check_instance_has_enabled_flow(config=config)

        instance_config = config.auth_config.get_or_create_instance(
            instance_name=defined_instance
        )
        OAuthClient(config, defined_instance).oauth_process(
            token_name=token_name, lifetime=lifetime, login_path=login_path
        )
    return 0



import codecs
import concurrent.futures
import logging
import re
from ast import literal_eval
from concurrent.futures import Future
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Iterable, List, NamedTuple, Optional, Set, Tuple

import charset_normalizer
import click
from pygitguardian import GGClient
from pygitguardian.config import DOCUMENT_SIZE_THRESHOLD_BYTES, MULTI_DOCUMENT_LIMIT
from pygitguardian.models import Detail, ScanResult

from ggshield.core.cache import Cache
from ggshield.core.extra_headers import get_headers
from ggshield.core.filter import (
    is_filepath_excluded,
    remove_ignored_from_result,
    remove_results_from_ignore_detectors,
)
from ggshield.core.git_shell import GIT_PATH, shell
from ggshield.core.text_utils import STYLE, display_error, format_text, pluralize
from ggshield.core.types import IgnoredMatch
from ggshield.core.utils import REGEX_HEADER_INFO, Filemode, ScanContext

from ..iac.models import IaCScanResult


logger = logging.getLogger(__name__)

_RX_HEADER_LINE_SEPARATOR = re.compile("[\n\0]:", re.MULTILINE)


def _parse_patch_header_line(line: str) -> Tuple[str, Filemode]:
    """
    Parse a file line in the raw patch header, returns a tuple of filename, filemode

    See https://github.com/git/git/blob/master/Documentation/diff-format.txt for details
    on the format.
    """

    prefix, name, *rest = line.rstrip("\0").split("\0")

    if rest:
        # If the line has a new name, we want to use it
        name = rest[0]

    # for a non-merge commit, prefix is
    # :old_perm new_perm old_sha new_sha status_and_score
    #
    # for a 2 parent commit, prefix is
    # ::old_perm1 old_perm2 new_perm old_sha1 old_sha2 new_sha status_and_score
    #
    # We can ignore most of it, because we only care about the status.
    #
    # status_and_score is one or more status letters, followed by an optional numerical
    # score. We can ignore the score, but we need to check the status letters.
    status = prefix.rsplit(" ", 1)[-1].rstrip("0123456789")

    # There is one status letter per commit parent. In the case of a non-merge commit
    # the situation is simple: there is only one letter.
    # In the case of a merge commit we must look at all letters: if one parent is marked
    # as D(eleted) and the other as M(odified) then we use MODIFY as filemode because
    # the end result contains modifications. To ensure this, the order of the `if` below
    # matters.

    if "M" in status:  # modify
        return name, Filemode.MODIFY
    elif "C" in status:  # copy
        return name, Filemode.NEW
    elif "A" in status:  # add
        return name, Filemode.NEW
    elif "T" in status:  # type change
        return name, Filemode.NEW
    elif "R" in status:  # rename
        return name, Filemode.RENAME
    elif "D" in status:  # delete
        return name, Filemode.DELETE
    else:
        raise ValueError(f"Can't parse header line {line}: unknown status {status}")


def _parse_patch_header(header: str) -> Iterable[Tuple[str, Filemode]]:
    """
    Parse the header of a raw patch, generated with -z --raw
    """

    if header[0] == ":":
        # If the patch has been generated by `git diff` and not by `git show` then
        # there is no commit info and message, add a blank line to simulate commit info
        # otherwise the split below is going to skip the first file of the patch.
        header = "\n" + header

    # First item returned by split() contains commit info and message, skip it
    for line in _RX_HEADER_LINE_SEPARATOR.split(header)[1:]:
        yield _parse_patch_header_line(f":{line}")


class PatchParseError(Exception):
    """
    Raised by Commit.get_files() if it fails to parse its patch.
    """

    pass


class Result(NamedTuple):
    """
    Return model for a scan which zips the information
    between the Scan result and its input content.
    """

    content: str  # Text content scanned
    filemode: Filemode  # Filemode (useful for commits)
    filename: str  # Filename of content scanned
    scan: ScanResult  # Result of content scan


class Error(NamedTuple):
    files: List[Tuple[str, Filemode]]
    description: str  # Description of the error


@dataclass(frozen=True)
class Results:
    """
    Return model for a scan with the results and errors of the scan

    Not a NamedTuple like the others because it causes mypy 0.961 to crash on the
    `from_exception()` method (!)

    Similar crash: https://github.com/python/mypy/issues/12629
    """

    results: List[Result]
    errors: List[Error]

    @staticmethod
    def from_exception(exc: Exception) -> "Results":
        """Create a Results representing a failure"""
        error = Error(files=[], description=str(exc))
        return Results(results=[], errors=[error])


class ScanCollection(NamedTuple):
    id: str
    type: str
    results: Optional[Results] = None
    scans: Optional[List["ScanCollection"]] = None  # type: ignore[misc]
    iac_result: Optional[IaCScanResult] = None
    optional_header: Optional[str] = None  # To be printed in Text Output
    extra_info: Optional[Dict[str, str]] = None  # To be included in JSON Output

    @property
    def scans_with_results(self) -> List["ScanCollection"]:
        if self.scans:
            return [scan for scan in self.scans if scan.results]
        return []

    @property
    def has_iac_result(self) -> bool:
        return bool(self.iac_result and self.iac_result.entities_with_incidents)

    @property
    def has_results(self) -> bool:
        return bool(self.results and self.results.results)

    def get_all_results(self) -> Iterable[Result]:
        """Returns an iterable on all results and sub-scan results"""
        if self.results:
            yield from self.results.results
        if self.scans:
            for scan in self.scans:
                yield from scan.results.results


class File:
    """Class representing a simple file."""

    def __init__(self, document: Optional[str], filename: str):
        self._document = document
        self.filename = filename
        self.filemode = Filemode.FILE

    def relative_to(self, root_path: Path) -> "File":
        return File(self._document, str(Path(self.filename).relative_to(root_path)))

    @property
    def document(self) -> str:
        self.prepare()
        assert self._document is not None
        return self._document

    def prepare(self) -> None:
        """Ensures self._document has been decoded"""
        if self._document is not None:
            return
        with open(self.filename, "rb") as f:
            self._document = File._decode_bytes(f.read(), self.filename)

    @staticmethod
    def from_bytes(raw_document: bytes, filename: str) -> "File":
        """Creates a File instance for a raw document. Document is decoded immediately."""
        document = File._decode_bytes(raw_document, filename)
        return File(document, filename)

    @staticmethod
    def from_path(filename: str) -> "File":
        """Creates a File instance for a file. Content is *not* read immediately."""
        return File(None, filename)

    @staticmethod
    def _decode_bytes(raw_document: bytes, filename: str) -> str:
        """Low level function to decode bytes, tries hard to find the correct encoding.
        For now it returns an empty string if the document could not be decoded"""
        result = charset_normalizer.from_bytes(raw_document).best()
        if result is None:
            # This means we were not able to detect the encoding. Report it using logging for now
            # TODO: we should report this in the output
            logger.warning("Skipping %s, can't detect encoding", filename)
            return ""

        # Special case for utf_8 + BOM: `bytes.decode()` does not skip the BOM, so do it
        # ourselves
        if result.encoding == "utf_8" and raw_document.startswith(codecs.BOM_UTF8):
            raw_document = raw_document[len(codecs.BOM_UTF8) :]
        return raw_document.decode(result.encoding, errors="replace")

    def __repr__(self) -> str:
        return f"<File filename={self.filename} filemode={self.filemode}>"

    def has_extensions(self, extensions: Set[str]) -> bool:
        """Returns True iff the file has one of the given extensions."""
        file_extensions = Path(self.filename).suffixes
        return any(ext in extensions for ext in file_extensions)


class CommitFile(File):
    """Class representing a commit file."""

    def __init__(self, document: str, filename: str, filemode: Filemode):
        super().__init__(document, filename)
        self.filemode = filemode


class Files:
    """
    Files is a list of files. Useful for directory scanning.
    """

    def __init__(self, files: List[File]):
        self._files = files

    @property
    def files(self) -> List[File]:
        """The list of files owned by this instance. The same filename can appear twice,
        in case of a merge commit."""
        return self._files

    @property
    def filenames(self) -> List[str]:
        """Convenience property to list filenames in the same order as files"""
        return [x.filename for x in self.files]

    def __repr__(self) -> str:
        return f"<Files files={self.files}>"

    def apply_filter(self, filter_func: Callable[[File], bool]) -> "Files":
        return Files([file for file in self.files if filter_func(file)])

    def relative_to(self, root_path: Path) -> "Files":
        return Files([file.relative_to(root_path) for file in self.files])

    def scan(
        self,
        client: GGClient,
        cache: Cache,
        matches_ignore: Iterable[IgnoredMatch],
        scan_context: ScanContext,
        ignored_detectors: Optional[Set[str]] = None,
        on_file_chunk_scanned: Callable[[List[File]], None] = lambda chunk: None,
        scan_threads: int = 4,
    ) -> Results:
        logger.debug("self=%s command_id=%s", self, scan_context.command_id)
        scanner = Scanner(
            client, cache, matches_ignore, scan_context, ignored_detectors
        )
        return scanner.scan(self.files, on_file_chunk_scanned, scan_threads)


class Scanner:
    def __init__(
        self,
        client: GGClient,
        cache: Cache,
        matches_ignore: Iterable[IgnoredMatch],
        scan_context: ScanContext,
        ignored_detectors: Optional[Set[str]] = None,
    ):
        self.client = client
        self.cache = cache
        self.matches_ignore = matches_ignore
        self.ignored_detectors = ignored_detectors
        self.headers = get_headers(scan_context)

    def _scan_chunk(
        self, executor: concurrent.futures.ThreadPoolExecutor, chunk: List[File]
    ) -> Future:
        """
        Sends a chunk of files to scan to the API
        """
        # `documents` is a version of `chunk` suitable for `GGClient.multi_content_scan()`
        documents = [{"document": x.document, "filename": x.filename} for x in chunk]
        return executor.submit(
            self.client.multi_content_scan,
            documents,
            self.headers,
        )

    def _start_scans(
        self, executor: concurrent.futures.ThreadPoolExecutor, files: Iterable[File]
    ) -> Tuple[Dict[Future, List[File]], List[File]]:
        """
        Start all scans, return a tuple containing:
        - a mapping of future to the list of files it is scanning
        - a list of files which we did not send to scan because we could not decode them
        """
        chunks_for_futures = {}
        skipped_chunk = []

        chunk: List[File] = []
        for file in files:
            if file.document:
                chunk.append(file)
                if len(chunk) == MULTI_DOCUMENT_LIMIT:
                    future = self._scan_chunk(executor, chunk)
                    chunks_for_futures[future] = chunk
                    chunk = []
            else:
                skipped_chunk.append(file)
        if chunk:
            future = self._scan_chunk(executor, chunk)
            chunks_for_futures[future] = chunk
        return chunks_for_futures, skipped_chunk

    def _collect_results(
        self,
        chunks_for_futures: Dict[Future, List[File]],
        on_file_chunk_scanned: Callable[[List[File]], None],
    ) -> Results:
        """
        Receive scans as they complete, report progress and collect them and return
        a Results.
        """
        self.cache.purge()

        results = []
        errors = []
        for future in concurrent.futures.as_completed(chunks_for_futures):
            chunk = chunks_for_futures[future]
            on_file_chunk_scanned(chunk)

            exception = future.exception()
            if exception is None:
                scan = future.result()
            else:
                scan = Detail(detail=str(exception))
                errors.append(
                    Error(
                        files=[(file.filename, file.filemode) for file in chunk],
                        description=scan.detail,
                    )
                )

            if not scan.success:
                handle_scan_chunk_error(scan, chunk)
                continue

            for file, scanned in zip(chunk, scan.scan_results):
                remove_ignored_from_result(scanned, self.matches_ignore)
                remove_results_from_ignore_detectors(scanned, self.ignored_detectors)
                if scanned.has_policy_breaks:
                    for policy_break in scanned.policy_breaks:
                        self.cache.add_found_policy_break(policy_break, file.filename)
                    results.append(
                        Result(
                            content=file.document,
                            scan=scanned,
                            filemode=file.filemode,
                            filename=file.filename,
                        )
                    )

        self.cache.save()
        return Results(results=results, errors=errors)

    def scan(
        self,
        files: Iterable[File],
        on_file_chunk_scanned: Callable[[List[File]], None],
        scan_threads: int = 4,
    ) -> Results:

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=scan_threads, thread_name_prefix="content_scan"
        ) as executor:
            chunks_for_futures, skipped_chunk = self._start_scans(executor, files)

            if skipped_chunk:
                # Report skipped_chunk as progress. If we only report files really sent
                # for scanning then the progress bar won't reach 100%.
                on_file_chunk_scanned(skipped_chunk)

            return self._collect_results(chunks_for_futures, on_file_chunk_scanned)


class CommitInformation(NamedTuple):
    author: str
    email: str
    date: str


def _parse_patch(
    patch: str, exclusion_regexes: Set[re.Pattern]
) -> Iterable[CommitFile]:
    """
    Parse the patch generated with `git show` (or `git diff`)

    If the patch represents a merge commit, then `patch` actually contains multiple
    commits, one per parent, because we call `git show` with the `-m` option to force it
    to generate one single-parent commit per parent. This makes later code simpler and
    ensures we see *all* the changes.
    """
    for commit in patch.split("\0commit "):
        tokens = commit.split("\0diff ", 1)
        if len(tokens) == 1:
            # No diff, carry on to next commit
            continue
        header, rest = tokens

        names_and_modes = _parse_patch_header(header)

        diffs = re.split(r"^diff ", rest, flags=re.MULTILINE)
        for (filename, filemode), diff in zip(names_and_modes, diffs):
            if is_filepath_excluded(filename, exclusion_regexes):
                continue

            # extract document from diff: we must skip diff extended headers
            # (lines like "old mode 100644", "--- a/foo", "+++ b/foo"...)
            try:
                end_of_headers = diff.index("\n@@")
            except ValueError:
                # No content
                continue
            # +1 because we searched for the '\n'
            document = diff[end_of_headers + 1 :]

            file_size = len(document.encode("utf-8"))
            if file_size > DOCUMENT_SIZE_THRESHOLD_BYTES * 0.90:
                continue

            if document:
                yield CommitFile(document, filename, filemode)


class Commit(Files):
    """
    Commit represents a commit which is a list of commit files.
    """

    def __init__(
        self, sha: Optional[str] = None, exclusion_regexes: Set[re.Pattern] = set()
    ):
        self.sha = sha
        self._patch: Optional[str] = None
        self._files = []
        self.exclusion_regexes = exclusion_regexes
        self._info: Optional[CommitInformation] = None

    @property
    def info(self) -> CommitInformation:
        if self._info is None:
            m = REGEX_HEADER_INFO.search(self.patch)

            if m is None:
                self._info = CommitInformation("unknown", "", "")
            else:
                self._info = CommitInformation(**m.groupdict())

        return self._info

    @property
    def optional_header(self) -> str:
        """Return the formatted patch."""
        return (
            format_text(f"\ncommit {self.sha}\n", STYLE["commit_info"])
            + f"Author: {self.info.author} <{self.info.email}>\n"
            + f"Date: {self.info.date}\n"
        )

    @property
    def patch(self) -> str:
        """Get the change patch for the commit."""
        if self._patch is None:
            common_args = [
                "--raw",  # shows a header with the files touched by the commit
                "-z",  # separate file names in the raw header with \0
                "--patch",  # force output of the diff (--raw disables it)
                "-m",  # split multi-parent (aka merge) commits into several one-parent commits
            ]
            if self.sha:
                self._patch = shell([GIT_PATH, "show", self.sha] + common_args)
            else:
                self._patch = shell([GIT_PATH, "diff", "--cached"] + common_args)

        return self._patch

    @property
    def files(self) -> List[File]:
        if not self._files:
            self._files = list(self.get_files())

        return self._files

    def get_files(self) -> Iterable[CommitFile]:
        """
        Parse the patch into files and extract the changes for each one of them.

        See tests/data/patches for examples
        """
        try:
            yield from _parse_patch(self.patch, self.exclusion_regexes)
        except Exception as exc:
            raise PatchParseError(f"Could not parse patch (sha: {self.sha}): {exc}")

    def __repr__(self) -> str:
        return f"<Commit sha={self.sha} files={self.files}>"


def handle_scan_chunk_error(detail: Detail, chunk: List[File]) -> None:
    logger.error("status_code=%d detail=%s", detail.status_code, detail.detail)
    if detail.status_code == 401:
        raise click.UsageError(detail.detail)

    details = None

    display_error("\nError scanning. Results may be incomplete.")
    try:
        # try to load as list of dicts to get per file details
        details = literal_eval(detail.detail)
    except Exception:
        pass

    if isinstance(details, list) and details:
        # if the details had per file details
        display_error(
            f"Add the following {pluralize('file', len(details))}"
            " to your paths-ignore:"
        )
        for i, inner_detail in enumerate(details):
            if inner_detail:
                click.echo(
                    f"- {format_text(chunk[i].filename, STYLE['filename'])}:"
                    f" {str(inner_detail)}",
                    err=True,
                )
        return
    else:
        # if the details had a request error
        filenames = ", ".join([file.filename for file in chunk])
        display_error(
            "The following chunk is affected:\n"
            f"{format_text(filenames, STYLE['filename'])}"
        )

        display_error(str(detail))



import os
from unittest import mock

import click.exceptions
import pytest

from ggshield.cmd.main import main


CRASH_LOG_ENV = {"GITGUARDIAN_CRASH_LOG": "true"}


def test_syntax_error_no_crash_log():
    with pytest.raises(SystemExit) as e:
        main(args=["foo"])
    assert e.value.code != 0


@mock.patch.dict(os.environ, CRASH_LOG_ENV)
def test_syntax_error_crash_log():
    with pytest.raises(click.exceptions.UsageError):
        main(args=["foo"])



import concurrent.futures
import os
import re
import sys
from contextlib import contextmanager
from typing import Iterable, Iterator, List, Optional, Set

import click
from pygitguardian import GGClient

from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.constants import MAX_WORKERS
from ggshield.core.git_shell import get_list_commit_SHA, is_git_dir
from ggshield.core.text_utils import STYLE, display_error, format_text
from ggshield.core.types import IgnoredMatch
from ggshield.core.utils import ScanContext, handle_exception
from ggshield.output import OutputHandler
from ggshield.scan import Commit, Results, ScanCollection


# We add a maximal value to avoid silently consuming all threads on powerful machines
SCAN_THREADS = 4


@contextmanager
def cd(newdir: str) -> Iterator[None]:
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def scan_repo_path(
    client: GGClient,
    cache: Cache,
    output_handler: OutputHandler,
    config: Config,
    scan_context: ScanContext,
    repo_path: str,
) -> int:  # pragma: no cover
    try:
        if not is_git_dir(repo_path):
            raise click.ClickException(f"{repo_path} is not a git repository")

        with cd(repo_path):
            return scan_commit_range(
                client=client,
                cache=cache,
                commit_list=get_list_commit_SHA("--all"),
                output_handler=output_handler,
                exclusion_regexes=set(),
                matches_ignore=config.secret.ignored_matches,
                scan_context=scan_context,
                ignored_detectors=config.secret.ignored_detectors,
            )
    except Exception as error:
        return handle_exception(error, config.verbose)


def scan_commit(
    commit: Commit,
    client: GGClient,
    cache: Cache,
    matches_ignore: Iterable[IgnoredMatch],
    scan_context: ScanContext,
    ignored_detectors: Optional[Set[str]] = None,
) -> ScanCollection:  # pragma: no cover
    try:
        results = commit.scan(
            client=client,
            cache=cache,
            matches_ignore=matches_ignore,
            scan_context=scan_context,
            ignored_detectors=ignored_detectors,
            scan_threads=SCAN_THREADS,
        )
    except Exception as exc:
        results = Results.from_exception(exc)

    return ScanCollection(
        commit.sha or "unknown",
        type="commit",
        results=results,
        optional_header=commit.optional_header,
        extra_info=commit.info._asdict(),
    )


def scan_commit_range(
    client: GGClient,
    cache: Cache,
    commit_list: List[str],
    output_handler: OutputHandler,
    exclusion_regexes: Set[re.Pattern],
    matches_ignore: Iterable[IgnoredMatch],
    scan_context: ScanContext,
    ignored_detectors: Optional[Set[str]] = None,
) -> int:  # pragma: no cover
    """
    Scan every commit in a range.

    :param client: Public Scanning API client
    :param commit_list: List of commits sha to scan
    :param verbose: Display successful scan's message
    """

    with concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:

        future_to_process = [
            executor.submit(
                scan_commit,
                Commit(sha, exclusion_regexes),
                client,
                cache,
                matches_ignore,
                scan_context,
                ignored_detectors,
            )
            for sha in commit_list
        ]

        scans: List[ScanCollection] = []
        with click.progressbar(
            iterable=concurrent.futures.as_completed(future_to_process),
            length=len(future_to_process),
            label=format_text("Scanning Commits", STYLE["progress"]),
            file=sys.stderr,
        ) as completed_futures:
            for future in completed_futures:
                scan_collection = future.result()
                if scan_collection.results and scan_collection.results.errors:
                    for error in scan_collection.results.errors:
                        # Prefix with `\n` since we are in the middle of a progress bar
                        display_error(f"\n{error.description}")
                scans.append(scan_collection)

        return_code = output_handler.process_scan(
            ScanCollection(id=scan_context.command_id, type="commit-range", scans=scans)
        )
    return return_code



from dataclasses import dataclass

import marshmallow_dataclass
from pygitguardian.models import Base, BaseSchema


@dataclass
class IaCVulnerability(Base):
    policy: str
    policy_id: str
    line_end: int
    line_start: int
    description: str
    documentation_url: str
    component: str = ""
    severity: str = ""


IaCVulnerabilitySchema = marshmallow_dataclass.class_schema(
    IaCVulnerability, BaseSchema
)



from dataclasses import dataclass, field
from typing import List

import marshmallow_dataclass
from pygitguardian.models import Base, BaseSchema

from ggshield.iac.models.iac_file_result import IaCFileResult


@dataclass
class IaCScanResult(Base):
    id: str = ""
    type: str = ""
    iac_engine_version: str = ""
    entities_with_incidents: List[IaCFileResult] = field(default_factory=list)


IaCScanResultSchema = marshmallow_dataclass.class_schema(IaCScanResult, BaseSchema)



import json
import re
from pathlib import Path
from typing import Any, Dict

import voluptuous.validators as validators
from click.testing import CliRunner, Result
from pytest_voluptuous import S

from ggshield.cmd.main import cli
from tests.conftest import (
    _IAC_MULTIPLE_VULNERABILITIES,
    _IAC_NO_VULNERABILITIES,
    _IAC_SINGLE_VULNERABILITY,
    my_vcr,
)


INCIDENT_SCHEMA = validators.Schema(
    {
        "policy": str,
        "policy_id": validators.Match(r"^GG_IAC_\d{4}$"),
        "line_end": int,
        "line_start": int,
        "description": str,
        "documentation_url": validators.All(str, validators.Match(r"^https://")),
        "component": str,
        "severity": validators.Any("LOW", "MEDIUM", "HIGH", "CRITICAL"),
    }
)


@my_vcr.use_cassette("test_iac_scan_single_vulnerability")
def test_display_single_vulnerabilities(cli_fs_runner: CliRunner):
    Path("tmp/").mkdir(exist_ok=True)
    Path("tmp/iac_file_single_vulnerability.tf").write_text(_IAC_SINGLE_VULNERABILITY)

    result = cli_fs_runner.invoke(
        cli,
        [
            "iac",
            "scan",
            "--json",
            "tmp",
        ],
    )

    json_result = load_json(result)
    assert_iac_version_displayed(json_result, 1)
    assert_file_single_vulnerability_displayed(json_result)


@my_vcr.use_cassette("test_iac_scan_multiple_vulnerabilities")
def test_display_multiple_vulnerabilities(cli_fs_runner: CliRunner):
    Path("tmp/").mkdir(exist_ok=True)
    Path("tmp/iac_file_multiple_vulnerabilities.tf").write_text(
        _IAC_MULTIPLE_VULNERABILITIES
    )

    result = cli_fs_runner.invoke(
        cli,
        [
            "iac",
            "scan",
            "--json",
            "tmp",
        ],
    )

    json_result = load_json(result)
    assert_iac_version_displayed(json_result, 2)
    assert_file_multiple_vulnerabilities_displayed(json_result)


@my_vcr.use_cassette("test_iac_scan_no_vulnerabilities")
def test_display_no_vulnerability(cli_fs_runner: CliRunner):
    Path("tmp/").mkdir(exist_ok=True)
    Path("tmp/iac_file_no_vulnerabilities.tf").write_text(_IAC_NO_VULNERABILITIES)

    result = cli_fs_runner.invoke(
        cli,
        [
            "iac",
            "scan",
            "--json",
            "tmp",
        ],
    )

    json_result = load_json(result)
    assert_iac_version_displayed(json_result, 0)
    assert len(json_result["entities_with_incidents"]) == 0


@my_vcr.use_cassette("test_iac_scan_multiple_files")
def test_display_multiple_files(cli_fs_runner: CliRunner):
    Path("tmp/").mkdir(exist_ok=True)
    Path("tmp/iac_file_single_vulnerability.tf").write_text(_IAC_SINGLE_VULNERABILITY)
    Path("tmp/iac_file_multiple_vulnerabilities.tf").write_text(
        _IAC_MULTIPLE_VULNERABILITIES
    )
    Path("tmp/iac_file_no_vulnerabilities.tf").write_text(_IAC_NO_VULNERABILITIES)

    result = cli_fs_runner.invoke(
        cli,
        [
            "iac",
            "scan",
            "--json",
            "tmp",
        ],
    )

    json_result = load_json(result)
    assert_iac_version_displayed(json_result, 3)
    assert_file_single_vulnerability_displayed(json_result)
    assert_file_multiple_vulnerabilities_displayed(json_result)


def load_json(result: Result) -> Dict[str, Any]:
    return json.loads(result.stdout)


def assert_iac_version_displayed(json_result: Dict[str, Any], total_incidents: int):
    assert re.match(r"\d\.\d{1,3}\.\d", json_result["iac_engine_version"])
    assert json_result["type"] == "path_scan"
    assert json_result["total_incidents"] == total_incidents


def assert_file_single_vulnerability_displayed(json_result: Dict[str, Any]):
    file_result = [
        file_result
        for file_result in json_result["entities_with_incidents"]
        if file_result["filename"] == "iac_file_single_vulnerability.tf"
    ]
    assert len(file_result) == 1
    assert (
        S(
            {
                "filename": str,
                "incidents": validators.All(
                    [INCIDENT_SCHEMA], validators.Length(min=1, max=1)
                ),
                "total_incidents": 1,
            }
        )
        == file_result[0]
    )
    assert file_result[0]["incidents"][0]["policy_id"] == "GG_IAC_0001"


def assert_file_multiple_vulnerabilities_displayed(json_result: Dict[str, Any]):
    file_result = [
        file_result
        for file_result in json_result["entities_with_incidents"]
        if file_result["filename"] == "iac_file_multiple_vulnerabilities.tf"
    ]
    assert len(file_result) == 1
    assert (
        S(
            {
                "filename": str,
                "incidents": validators.All(
                    [INCIDENT_SCHEMA], validators.Length(min=2, max=2)
                ),
                "total_incidents": 2,
            }
        )
        == file_result[0]
    )
    assert {incident["policy_id"] for incident in file_result[0]["incidents"]} == {
        "GG_IAC_0002",
        "GG_IAC_0003",
    }



import os
import sys
from copy import deepcopy

import pytest

from ggshield.core.config import Config
from ggshield.core.config.auth_config import (
    AuthConfigSchema,
    prepare_auth_config_dict_for_save,
)
from ggshield.core.config.utils import get_auth_config_filepath, replace_in_keys
from tests.conftest import write_text, write_yaml
from tests.core.config.conftest import TEST_AUTH_CONFIG


@pytest.fixture(autouse=True)
def env_vars(monkeypatch):
    monkeypatch.setenv("GITGUARDIAN_API_URL", "https://api.gitguardian.com")


@pytest.mark.usefixtures("isolated_fs")
class TestAuthConfig:
    def test_load(self):
        """
        GIVEN a default auth config
        WHEN loading the config
        THEN when serializing it again, it matches the data.
        """
        write_yaml(get_auth_config_filepath(), TEST_AUTH_CONFIG)

        config = Config()

        assert config.instances[0].account.token_name == "my_token"
        assert config.instances[0].default_token_lifetime == 1
        assert config.auth_config.default_token_lifetime == 2

        config_data = AuthConfigSchema().dump(config.auth_config)
        config_data = prepare_auth_config_dict_for_save(config_data)
        replace_in_keys(config_data, old_char="_", new_char="-")
        assert config_data == TEST_AUTH_CONFIG

    @pytest.mark.parametrize("n", [0, 2])
    def test_no_account(self, n):
        """
        GIVEN an auth config with a instance with 0 or more than 1 accounts
        WHEN loading the AuthConfig
        THEN it raises
        """
        raw_config = deepcopy(TEST_AUTH_CONFIG)
        raw_config["instances"][0]["accounts"] = (
            raw_config["instances"][0]["accounts"] * n
        )
        write_yaml(get_auth_config_filepath(), raw_config)

        with pytest.raises(
            AssertionError,
            match="Each GitGuardian instance should have exactly one account",
        ):
            Config()

    def test_invalid_format(self, capsys):
        """
        GIVEN an auth config file with invalid content
        WHEN loading AuthConfig
        THEN it raises
        """
        write_text(get_auth_config_filepath(), "Not a:\nyaml file.\n")

        Config()
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)

        assert f"Parsing error while reading {get_auth_config_filepath()}:" in err

    def test_token_not_expiring(self):
        """
        GIVEN an auth config file with a token never expiring
        WHEN loading the AuthConfig
        THEN it works
        """
        raw_config = deepcopy(TEST_AUTH_CONFIG)
        raw_config["instances"][0]["accounts"][0]["expire-at"] = None
        write_yaml(get_auth_config_filepath(), raw_config)

        config = Config()

        assert config.instances[0].account.expire_at is None

    def test_update(self):
        """
        GIVEN -
        WHEN modifying the default config
        THEN it's not persisted until .save() is called
        """
        config = Config()
        config.instance = "custom"

        assert Config().instance != "custom"

        config.save()

        assert Config().instance == "custom"

    def test_load_file_not_existing(self):
        """
        GIVEN the auth config file not existing
        WHEN loading the config
        THEN it works and has the default configuration
        """
        config = Config()

        assert config.instance_name == "https://dashboard.gitguardian.com"
        assert config.instances == []

    def test_save_file_not_existing(self):
        """
        GIVEN a config object and the auth config file not existing
        WHEN saving the config
        THEN it works and when loading the config again it has the correct values
        """
        config = Config()
        try:
            os.remove(get_auth_config_filepath())
        except FileNotFoundError:
            pass

        config.instance = "custom"
        config.save()
        updated_config = Config()

        assert updated_config.instance == "custom"

    def test_timezone_aware_expired(self):
        """
        GIVEN a config with a configured instance
        WHEN loading the config
        THEN the instance expiration date is timezone aware
        """
        write_yaml(get_auth_config_filepath(), TEST_AUTH_CONFIG)
        config = Config()
        assert config.instances[0].account.expire_at.tzinfo is not None



import re
import subprocess
from pathlib import Path
from typing import Dict
from unittest.mock import patch

import click
import pytest

from ggshield.scan.docker import (
    InvalidDockerArchiveException,
    _get_config,
    _should_scan_layer,
    docker_pull_image,
    docker_save_to_tmp,
    get_files_from_docker_archive,
)
from tests.conftest import DATA_PATH


DOCKER_EXAMPLE_PATH = DATA_PATH / "docker-example.tar.xz"
DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH = (
    DATA_PATH / "docker-incomplete-manifest-example.tar.xz"
)


class ManifestMock:
    def read(self, amount: int = None) -> bytes:
        return '[{"Config": "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4.json"}]'.encode(  # noqa: E501
            "utf-8"
        )


class TarMock:
    def __init__(self, members: Dict[str, str], *args, **kwargs):
        self.members = members

    def extractfile(self, member: str):
        if "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4" in member:
            return None
        return self.members.get(member, None)

    def getmember(self, member: str):
        return member if self.members.get(member, None) else None


class TestDockerScan:
    @pytest.mark.parametrize(
        ["op", "want"],
        [
            pytest.param(
                "/bin/sh -c #(nop) COPY dir:xxx in / ",
                True,
            ),
            pytest.param("/bin/sh -c #(nop) ADD dir:xxx in / ", True),
            pytest.param(
                '/bin/sh -c #(nop)  CMD ["/usr/bin/bash"',
                False,
            ),
        ],
    )
    def test_should_scan_layer(self, op: str, want: bool):
        assert _should_scan_layer({"created_by": op}) is want

    @pytest.mark.parametrize(
        ["members", "match"],
        [
            pytest.param({}, "No manifest file found."),
            pytest.param(
                {"manifest.json": ManifestMock()},
                "No config file found.",
            ),
            pytest.param(
                {
                    "manifest.json": ManifestMock(),
                    "8b907fee27ad927c595fcf873c8256796cab27e7a3fb4bf3952308a76ad791c4.json": "layer file",  # noqa: E501
                },  # noqa: E501
                "Config file could not be extracted.",
            ),
        ],
    )
    def test_get_config(self, members, match):
        tarfile = TarMock(members)
        with pytest.raises(InvalidDockerArchiveException, match=match):
            _get_config(tarfile)

    @pytest.mark.parametrize(
        "image_path", [DOCKER_EXAMPLE_PATH, DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH]
    )
    def test_get_files_from_docker_archive(self, image_path: Path):
        files = get_files_from_docker_archive(image_path)

        expected_files = {
            "Dockerfile or build-args": None,
            "64a345482d74ea1c0699988da4b4fe6cda54a2b0ad5da49853a9739f7a7e5bbc:/app/file_one": "Hello, I am the first file!\n",  # noqa: E501
            "2d185b802fb3c2e6458fe1ac98e027488cd6aedff2e3d05eb030029c1f24d60f:/app/file_three.sh": "echo Life is beautiful.\n",  # noqa: E501
            "2d185b802fb3c2e6458fe1ac98e027488cd6aedff2e3d05eb030029c1f24d60f:/app/file_two.py": """print("Hi! I'm the second file but I'm happy.")\n""",  # noqa: E501
        }

        assert set(files.filenames) == {str(file_path) for file_path in expected_files}

        file_dict = {x.filename: x for x in files.files}
        for file_path, expected_content in expected_files.items():
            file = file_dict[str(file_path)]
            assert expected_content is None or file.document == expected_content


DOCKER_TIMEOUT = 12


class TestDockerPull:
    def test_docker_pull_image_success(self):
        with patch("subprocess.run") as call:
            docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)
            call.assert_called_once_with(
                ["docker", "pull", "ggshield-non-existant"],
                check=True,
                timeout=DOCKER_TIMEOUT,
            )

    def test_docker_pull_image_non_exist(self):
        with patch(
            "subprocess.run", side_effect=subprocess.CalledProcessError(1, cmd=[])
        ):
            with pytest.raises(
                click.exceptions.ClickException,
                match='Image "ggshield-non-existant" not found',
            ):
                docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)

    def test_docker_pull_image_timeout(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.TimeoutExpired(cmd=[], timeout=DOCKER_TIMEOUT),
        ):
            with pytest.raises(
                click.exceptions.ClickException,
                match='docker pull ggshield-non-existant" timed out',
            ):
                docker_pull_image("ggshield-non-existant", DOCKER_TIMEOUT)


class TestDockerSave:
    TMP_ARCHIVE = Path("/tmp/as/archive.tar")

    def test_docker_save_image_success(self):
        with patch("subprocess.run") as call:
            docker_save_to_tmp(
                "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
            )
            call.assert_called_once_with(
                [
                    "docker",
                    "save",
                    "ggshield-non-existant",
                    "-o",
                    str(self.TMP_ARCHIVE),
                ],
                check=True,
                stderr=-1,
                timeout=DOCKER_TIMEOUT,
            )

    def test_docker_save_image_does_not_exist(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(
                1, cmd=[], stderr="reference does not exist".encode("utf-8")
            ),
        ):
            with pytest.raises(
                click.exceptions.ClickException,
                match='Image "ggshield-non-existant" not found',
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )

    def test_docker_save_image_need_pull(self):
        """
        GIVEN a Docker image we do not have locally
        WHEN we try to save it
        THEN we first pull it and then save it

        This test expects the following calls to `docker` commands:

        - docker save <image_name> -o <something>
          -> Fake failure
        - docker pull <image_name>
          -> Fake success
        - docker save <image_name> -o <something>
          -> Fake success
        """
        with patch(
            "subprocess.run",
            side_effect=[
                subprocess.CalledProcessError(
                    1, cmd=[], stderr="reference does not exist".encode("utf-8")
                ),
                None,
                None,
            ],
        ):
            docker_save_to_tmp(
                "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
            )

    def test_docker_save_image_fails(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(
                1, cmd=[], stderr="docker failed weirdly".encode("utf-8")
            ),
        ):
            with pytest.raises(
                click.exceptions.ClickException,
                match="Unable to save docker archive:",
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )

    def test_docker_save_image_timeout(self):
        with patch(
            "subprocess.run",
            side_effect=subprocess.TimeoutExpired(cmd=[], timeout=DOCKER_TIMEOUT),
        ):
            expected_msg = f'Command "docker save ggshield-non-existant -o {str(self.TMP_ARCHIVE)}" timed out'  # noqa: E501
            with pytest.raises(
                click.exceptions.ClickException,
                match=re.escape(expected_msg),
            ):
                docker_save_to_tmp(
                    "ggshield-non-existant", self.TMP_ARCHIVE, DOCKER_TIMEOUT
                )



from copy import deepcopy
from unittest import mock

import click
import pytest

from ggshield.core.utils import Filemode
from ggshield.output import TextOutputHandler
from ggshield.output.text.message import (
    _file_info_decoration,
    _file_info_default_decoration,
)
from ggshield.scan import Result
from ggshield.scan.scannable import Results, ScanCollection
from tests.conftest import (
    _MULTI_SECRET_ONE_LINE_PATCH,
    _MULTI_SECRET_ONE_LINE_PATCH_OVERLAY,
    _MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT,
    _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
    _MULTI_SECRET_TWO_LINES_PATCH,
    _MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
    _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_MULTILINE_PATCH,
    _SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_PATCH,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
)


@pytest.mark.parametrize(
    "show_secrets",
    [pytest.param(True, id="show_secrets"), pytest.param(False, id="hide_secrets")],
)
@pytest.mark.parametrize(
    "verbose",
    [pytest.param(True, id="verbose"), pytest.param(False, id="clip_long_lines")],
)
@pytest.mark.parametrize(
    "result_input",
    [
        pytest.param(
            Result(
                content=_SIMPLE_SECRET_PATCH,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_SIMPLE_SECRET_PATCH_SCAN_RESULT,
            ),
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result(
                content=_MULTI_SECRET_ONE_LINE_PATCH,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
            ),
            id="_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result(
                content=_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT,
            ),
            id="_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT",
        ),
        pytest.param(
            Result(
                content=_MULTI_SECRET_TWO_LINES_PATCH,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
            ),
            id="_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result(
                content=_SIMPLE_SECRET_MULTILINE_PATCH,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
            ),
            id="_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result(
                content=_ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
                filename="leak.txt",
                filemode=Filemode.NEW,
                scan=_ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
            ),
            id="_ONE_LINE_AND_MULTILINE_PATCH_CONTENT",
        ),
    ],
)
def test_leak_message(result_input, snapshot, show_secrets, verbose):
    # The text output includes the version of the secrets engine, but this version is
    # None until we make an API call. Since this test does not make any API call, set
    # the version to a fake value.
    with mock.patch("ggshield.output.text.message.VERSIONS") as VERSIONS:
        VERSIONS.secrets_engine_version = "3.14.159"

        output_handler = TextOutputHandler(show_secrets=show_secrets, verbose=verbose)

        # _process_scan_impl() modifies its ScanCollection arg(!), so make a copy of it
        new_result = deepcopy(result_input)

        output = output_handler._process_scan_impl(
            ScanCollection(
                id="scan",
                type="test",
                results=Results(results=[new_result], errors=[]),
                optional_header="> This is an example header",
            )
        )
    # Make output OS-independent, so that it can be safely compared to snapshots
    # regardless of the current OS:
    # - Remove colors because color codes are not the same on all OSes
    # - Replace any custom decoration with the default one
    output = click.unstyle(output).replace(
        _file_info_decoration(), _file_info_default_decoration()
    )

    snapshot.assert_match(output)



import re
import subprocess
from pathlib import Path

import pytest

from ggshield.iac.filter import get_iac_files_from_paths


FILE_NAMES = [
    "file1.txt",
    "file2.json",
    "file3.yaml",
    "file4.yml",
    "file5.jinja",
    "file6.py",
    "file7.py.schema",
    "file8.jinja.schema",
    "file9.tf",
    "filetfvars.anything",
    "dockerfile.txt",
]


def test_get_iac_files_from_paths(tmp_path):
    """
    GIVEN files added to the temp path
    WHEN calling get_iac_files_from_paths
    THEN it returns all iac files
    """
    tmp_paths = [str(tmp_path / filename) for filename in FILE_NAMES]
    for path in tmp_paths:
        Path(path).write_text("something")

    files = get_iac_files_from_paths(tmp_path, set(), True)
    assert len(files.files) == 10
    assert "file1.json" not in files.filenames
    assert "file2.json" in files.filenames


def test_get_iac_files_from_paths_excluded(tmp_path):
    """
    GIVEN files added to the temp path
    WHEN calling get_iac_files_from_paths with an excluded pattern
    THEN it returns all iac files expect the ones matching the excluded patterns
    """
    tmp_paths = [str(tmp_path / filename) for filename in FILE_NAMES]
    for path in tmp_paths:
        Path(path).write_text("something")

    files = get_iac_files_from_paths(tmp_path, {re.compile(r"file2")}, True)
    assert len(files.files) == 9
    assert "file2.json" not in files.filenames
    assert "file3.yaml" in files.filenames


@pytest.mark.parametrize("ignore_git", (False, True))
def test_get_iac_files_from_paths_ignore_git(tmp_path, ignore_git):
    """
    GIVEN files added to the temp path, as a git directory
    WHEN calling get_iac_files_from_paths with ignore_git
    THEN it returns all iac files added to git and not the ones mentioned in the
    .gitignore if ignore_git is True. Otherwise, it ignores the .gitignore
    """
    tmp_paths = [str(tmp_path / filename) for filename in FILE_NAMES]
    for path in tmp_paths:
        Path(path).write_text("something")
    Path(str(tmp_path / ".gitignore")).write_text("file2.json")
    tmp_path_str = str(tmp_path)
    subprocess.run(["git", "init"], cwd=tmp_path_str)
    subprocess.run(["git", "add", "."], cwd=tmp_path_str)

    # Assert the .git folder exists. files.files length assertion ensures it's not
    # included in the get_iac_files_from_paths response
    assert Path(str(tmp_path / ".git")).exists()

    files = get_iac_files_from_paths(tmp_path, set(), True, ignore_git)
    if ignore_git:
        assert len(files.files) == 10
        assert "file2.json" in files.filenames
    else:
        assert len(files.files) == 9
        assert "file2.json" not in files.filenames
        assert "file3.yaml" in files.filenames



from enum import Enum, auto
from typing import Any, Dict, NamedTuple, Optional, Union

import click


LINE_DISPLAY = {"file": "{} | ", "patch": "{} {} | "}

STYLE: Dict[str, Dict[str, Any]] = {
    "nb_secrets": {"fg": "bright_blue", "bold": True},
    "filename": {"fg": "bright_yellow", "bold": True},
    "commit_info": {"fg": "bright_yellow", "bold": False},
    "patch": {"fg": "white"},
    "key": {"fg": "bright_white", "bold": True},
    "secret": {"fg": "bright_red"},
    "error": {"fg": "red"},
    "no_secret": {"fg": "white"},
    "detector": {"fg": "bright_yellow", "bold": True},
    "policy": {"fg": "cyan", "bold": True},
    "ignore_sha": {"fg": "cyan"},
    "detector_line_start": {"fg": "cyan"},
    "line_count": {"fg": "white", "dim": True},
    "line_count_secret": {"fg": "yellow"},
    "progress": {"fg": "bright_yellow", "bold": False},
    "warning": {"fg": "yellow"},
}


class LineCategory(Enum):
    addition = auto()
    data = auto()
    deletion = auto()
    empty = auto()


class Line(NamedTuple):
    """
    Line object making easier to handle line
    by line display.

    - content: Content of the line
    - category: The line category [+|-| ] (addition, deletion, untouched)
    - pre_index: Line index (deletion for patches, line index for files)
    - post_index: Line index (addition for patches)
    """

    content: str
    category: Optional[LineCategory] = None
    pre_index: Optional[int] = None
    post_index: Optional[int] = None

    def build_line_count(self, padding: int, is_secret: bool = False) -> str:
        """Return the formatted line count."""
        line_count_style = (
            STYLE["line_count_secret"] if is_secret else STYLE["line_count"]
        )
        if self.category is not None and not isinstance(self.category, LineCategory):
            raise TypeError("line category invalid")

        # File
        if self.category == LineCategory.data:
            return LINE_DISPLAY["file"].format(
                format_text(
                    format_line_count(self.pre_index, padding), line_count_style
                )
            )

        # Patch
        pre_index = format_line_count(self.pre_index, padding)
        post_index = format_line_count(self.post_index, padding)

        if self.category == LineCategory.addition:
            pre_index = " " * padding

        elif self.category == LineCategory.deletion:
            post_index = " " * padding

        return LINE_DISPLAY["patch"].format(
            format_text(pre_index, line_count_style),
            format_text(post_index, line_count_style),
        )


def format_text(text: str, style: Dict[str, Any]) -> str:
    """Return the formatted text with the given style."""
    return click.style(
        text, fg=style["fg"], bold=style.get("bold", False), dim=style.get("dim", False)
    )


def pluralize(name: str, nb: int, plural: Union[str, None] = None) -> str:
    if nb == 1:
        return name
    return plural or (name + "s")


def format_line_count(line_count: Union[int, None], padding: int) -> str:
    """Return the padded line count."""
    if line_count is None:
        return " " * padding

    return " " * max(0, padding - len(str(line_count))) + str(line_count)


def display_error(msg: str) -> None:
    click.echo(format_text(msg, STYLE["error"]), err=True)


def display_warning(msg: str) -> None:
    click.echo(format_text(msg, STYLE["warning"]), err=True)


def display_info(msg: str, nl: bool = True) -> None:
    click.echo(msg, nl=nl, err=True)


_VALIDITY_TEXT_FOR_ID = {
    "unknown": "Unknown",
    # cannot_check is the old ID for secrets for which there are no checkers
    "cannot_check": "Cannot Check",
    "no_checker": "No Checker",
    "failed_to_check": "Failed to Check",
    "not_checked": "Not Checked",
    "invalid": "Invalid",
    "valid": "Valid",
}


def translate_validity(validity_id: Optional[str]) -> str:
    if validity_id is None:
        validity_id = "unknown"
    # If we don't have a text for the validity_id, return it as is. We assume the text
    # of the ID is more valuable than a generic "Unknown" string
    return _VALIDITY_TEXT_FOR_ID.get(validity_id, validity_id)



# Maintenance notes:
# - for readability, keep this set sorted
# - Guardians: keep this in sync with the server side
BINARY_EXTENSIONS = {
    "3g2",
    "3gp",
    "7z",
    "8svx",
    "a",
    "aa",
    "aac",
    "aax",
    "ace",
    "act",
    "adf",
    "afa",
    "aiff",
    "aml",
    "amr",
    "amv",
    "ape",
    "apk",
    "ar",
    "arc",
    "arj",
    "au",
    "avi",
    "awv",
    "b1",
    "b6z",
    "ba",
    "bh",
    "bin",
    "bmp",
    "bz2",
    "cab",
    "car",
    "cfs",
    "cpio",
    "cpt",
    "dar",
    "db",
    "dct",
    "dd",
    "dgc",
    "dll",
    "dmg",
    "drc",
    "dss",
    "dvf",
    "ear",
    "ecc",
    "efi",
    "exe",
    "f",
    "f4a",
    "f4b",
    "f4p",
    "f4v",
    "flac",
    "flv",
    "fpx",
    "gca",
    "gif",
    "gifv",
    "gsm",
    "gz",
    "ha",
    "hki",
    "ice",
    "icns",
    "ico",
    "ide",
    "iklax",
    "iso",
    "ivs",
    "j2c",
    "j2k",
    "jar",
    "jfif",
    "jif",
    "jp2",
    "jpeg",
    "jpg",
    "jpx",
    "kgb",
    "lbr",
    "lha",
    "lz",
    "lzh",
    "lzma",
    "lzo",
    "lzx",
    "m2ts",
    "m4a",
    "m4b",
    "m4p",
    "m4v",
    "mar",
    "mkv",
    "mmdb",
    "mmf",
    "mng",
    "mo",
    "mogg",
    "mov",
    "mp2",
    "mp3",
    "mp4",
    "mpc",
    "mpe",
    "mpeg",
    "mpg",
    "mpv",
    "msv",
    "mts",
    "mxf",
    "nsf",
    "nsv",
    "o",
    "oga",
    "ogg",
    "ogv",
    "opus",
    "pak",
    "paq6",
    "paq7",
    "paq8",
    "par",
    "par2",
    "partimg",
    "pcd",
    "pdf",
    "pea",
    "pim",
    "pit",
    "png",
    "pyc",
    "qda",
    "qt",
    "ra",
    "rar",
    "raw",
    "rev",
    "rk",
    "rm",
    "rmvb",
    "roq",
    "rz",
    "s7z",
    "sbx",
    "sda",
    "sea",
    "sen",
    "ser",
    "sfark",
    "sfx",
    "shar",
    "shk",
    "sit",
    "sitx",
    "sldasm",
    "slddrt",
    "slddrw",
    "sldprt",
    "sln",
    "sqlite",
    "sqx",
    "suo",
    "svg",
    "svi",
    "sz",
    "tar",
    "tbz2",
    "tgz",
    "tif",
    "tiff",
    "tlz",
    "tta",
    "twbw",
    "txz",
    "uc",
    "uc0",
    "uc1",
    "uc2",
    "uca",
    "ucn",
    "ue2",
    "uha",
    "ur2",
    "vob",
    "vox",
    "war",
    "wav",
    "webm",
    "wim",
    "wmv",
    "wv",
    "xar",
    "xcf",
    "xls",
    "xlsx",
    "xp3",
    "xz",
    "yuv",
    "yz1",
    "z",
    "zip",
    "zipx",
    "zoo",
    "zpaq",
    "zz",
}



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set

import pytest

from ggshield.core.config.utils import (
    remove_common_dict_items,
    replace_in_keys,
    update_from_other_instance,
)


def test_replace_in_keys():
    data = {"last-found-secrets": {"XXX"}}
    replace_in_keys(data, "-", "_")
    assert data == {"last_found_secrets": {"XXX"}}
    replace_in_keys(data, "_", "-")
    assert data == {"last-found-secrets": {"XXX"}}


@dataclass
class ExampleSubConfig:
    an_int: Optional[int] = None
    an_str: Optional[str] = None


@dataclass
class ExampleConfig:
    path: Optional[str] = None
    backup: bool = True
    subconf: ExampleSubConfig = field(default_factory=ExampleSubConfig)
    a_list: List[str] = field(default_factory=list)
    a_set: Set[str] = field(default_factory=set)


def test_update_from_other_instance():
    """
    GIVEN two dataclass instances
    WHEN update_from_other_instance(dst_conf, src_conf) is called
    THEN dst_conf is updated from src_conf fields
    """

    dst_conf = ExampleConfig(path="dst_path")
    dst_conf.subconf.an_int = 1
    dst_conf.a_list = ["i1", "i2"]
    dst_conf.a_set = {"i1", "i2"}

    src_conf = ExampleConfig(backup=False)
    src_conf.subconf.an_str = "src_subconf_str"
    src_conf.a_list = ["i2", "i3"]
    src_conf.a_set = {"i2", "i3"}

    update_from_other_instance(dst_conf, src_conf)

    assert dst_conf.path == "dst_path"
    assert dst_conf.backup is False
    assert dst_conf.subconf.an_int == 1
    assert dst_conf.subconf.an_str == "src_subconf_str"
    assert dst_conf.a_list == ["i1", "i2", "i2", "i3"]
    assert dst_conf.a_set == {"i1", "i2", "i3"}


@pytest.mark.parametrize(
    ["src", "reference", "expected"],
    [
        pytest.param(
            {"modified": 2, "same": 3},
            {"modified": 1, "same": 3, "absent": 4},
            {"modified": 2},
            id="simple",
        ),
        pytest.param(
            {"lst": [3, 4]},
            {"lst": [1, 2]},
            {"lst": [3, 4]},
            id="replace-list",
        ),
        pytest.param(
            {"lst1": [], "lst2": ["a"]},
            {"lst1": [], "lst2": ["a"]},
            {},
            id="identical-lists-are-removed",
        ),
        pytest.param(
            {"lst": []},
            {"lst": ["a"]},
            {"lst": []},
            id="empty-list-not-removed-if-ref-not-empty",
        ),
        pytest.param(
            {"outer": {"inner1": "modified", "inner3": "same"}},
            {"outer": {"inner1": "original", "inner2": "absent", "inner3": "same"}},
            {"outer": {"inner1": "modified"}},
            id="nested-merge",
        ),
        pytest.param(
            {"outer": {"inner1": "same"}},
            {"outer": {"inner1": "same", "inner2": "absent"}},
            {},
            id="nested-dict-is-removed",
        ),
    ],
)
def test_remove_common_dict_items(
    src: Dict[str, Any], reference: Dict[str, Any], expected: Dict[str, Any]
):
    result = remove_common_dict_items(src, reference)
    assert result == expected



import logging
import os
import sys
from typing import List, Optional, Tuple

import click

from ggshield.core.git_shell import check_git_dir, get_list_commit_SHA
from ggshield.core.utils import (
    EMPTY_SHA,
    EMPTY_TREE,
    ScanContext,
    ScanMode,
    handle_exception,
)
from ggshield.scan.repo import scan_commit_range


logger = logging.getLogger(__name__)


@click.command()
@click.argument("prepush_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def prepush_cmd(ctx: click.Context, prepush_args: List[str]) -> int:  # pragma: no cover
    """
    scan as a pre-push git hook.
    """
    config = ctx.obj["config"]

    local_commit, remote_commit = collect_from_precommit_env()
    if local_commit is None or remote_commit is None:
        local_commit, remote_commit = collect_from_stdin()

    if local_commit == EMPTY_SHA:
        click.echo("Deletion event or nothing to scan.", err=True)
        return 0

    if remote_commit == EMPTY_SHA:
        click.echo(
            f"New tree event. Scanning last {config.max_commits_for_hook} commits.",
            err=True,
        )
        before = EMPTY_TREE
        after = local_commit
        cmd_range = (
            f"--max-count={config.max_commits_for_hook+1} {EMPTY_TREE} {local_commit}"
        )
    else:
        before = remote_commit
        after = local_commit
        cmd_range = f"--max-count={config.max_commits_for_hook+1} {remote_commit}...{local_commit}"  # noqa

    commit_list = get_list_commit_SHA(cmd_range)

    if not commit_list:
        click.echo(
            "Unable to get commit range.\n"
            f"  before: {before}\n"
            f"  after: {after}\n"
            "Skipping pre-push hook\n",
            err=True,
        )
        return 0

    if len(commit_list) > config.max_commits_for_hook:
        click.echo(
            f"Too many commits. Scanning last {config.max_commits_for_hook} commits\n",
            err=True,
        )
        commit_list = commit_list[-config.max_commits_for_hook :]

    if config.verbose:
        click.echo(f"Commits to scan: {len(commit_list)}", err=True)

    try:
        check_git_dir()

        scan_context = ScanContext(
            scan_mode=ScanMode.PRE_PUSH,
            command_path=ctx.command_path,
        )

        return scan_commit_range(
            client=ctx.obj["client"],
            cache=ctx.obj["cache"],
            commit_list=commit_list,
            output_handler=ctx.obj["output_handler"],
            exclusion_regexes=ctx.obj["exclusion_regexes"],
            matches_ignore=config.secret.ignored_matches,
            scan_context=scan_context,
            ignored_detectors=config.secret.ignored_detectors,
        )
    except Exception as error:
        return handle_exception(error, config.verbose)


def collect_from_stdin() -> Tuple[str, str]:
    """
    Collect pre-commit variables from stdin.
    """
    prepush_input = sys.stdin.read().split()
    logger.debug("input=%s", prepush_input)
    if len(prepush_input) < 4:
        # Then it's either a tag or a deletion event
        local_commit = EMPTY_SHA
        remote_commit = EMPTY_SHA
    else:
        local_commit = prepush_input[1].strip()
        remote_commit = prepush_input[3].strip()

    logger.debug("refs=(%s, %s)", local_commit, remote_commit)
    return (local_commit, remote_commit)


def collect_from_precommit_env() -> Tuple[Optional[str], Optional[str]]:
    """
    Collect from pre-commit framework environment.
    """
    # pre-commit framework <2.2.0
    local_commit = os.getenv("PRE_COMMIT_SOURCE", None)
    remote_commit = os.getenv("PRE_COMMIT_ORIGIN", None)

    if local_commit is None or remote_commit is None:
        # pre-commit framework >=2.2.0
        local_commit = os.getenv("PRE_COMMIT_FROM_REF", None)
        remote_commit = os.getenv("PRE_COMMIT_TO_REF", None)

    logger.debug("refs=(%s, %s)", local_commit, remote_commit)
    return (local_commit, remote_commit)
